# /*##########################################################################
#
# Copyright (c) 2014-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
This module provides convenient classes and functions for OpenGL rendering.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "03/04/2017"


import numpy

from ...._glutils import gl


def buildFillMaskIndices(nIndices, dtype=None):
    """Returns triangle strip indices for rendering a filled polygon mask

    :param int nIndices: Number of points
    :param Union[numpy.dtype,None] dtype:
       If specified the dtype of the returned indices array
    :return: 1D array of indices constructing a triangle strip
    :rtype: numpy.ndarray
    """
    if dtype is None:
        if nIndices <= numpy.iinfo(numpy.uint16).max + 1:
            dtype = numpy.uint16
        else:
            dtype = numpy.uint32

    lastIndex = nIndices - 1
    splitIndex = lastIndex // 2 + 1
    indices = numpy.empty(nIndices, dtype=dtype)
    indices[::2] = numpy.arange(0, splitIndex, step=1, dtype=dtype)
    indices[1::2] = numpy.arange(lastIndex, splitIndex - 1, step=-1, dtype=dtype)
    return indices


class FilledShape2D(object):
    _NO_HATCH = 0
    _HATCH_STEP = 20

    def __init__(self, points, style="solid", color=(0.0, 0.0, 0.0, 1.0)):
        self.vertices = numpy.asarray(points, dtype=numpy.float32)
        self._indices = buildFillMaskIndices(len(self.vertices))

        tVertex = numpy.transpose(self.vertices)
        xMin, xMax = min(tVertex[0]), max(tVertex[0])
        yMin, yMax = min(tVertex[1]), max(tVertex[1])
        self.bboxVertices = numpy.array(
            ((xMin, yMin), (xMin, yMax), (xMax, yMin), (xMax, yMax)),
            dtype=numpy.float32,
        )
        self._xMin, self._xMax = xMin, xMax
        self._yMin, self._yMax = yMin, yMax

        self.style = style
        self.color = color

    def render(self, posAttrib, colorUnif, hatchStepUnif):
        assert self.style in ("hatch", "solid")
        gl.glUniform4f(colorUnif, *self.color)
        step = self._HATCH_STEP if self.style == "hatch" else self._NO_HATCH
        gl.glUniform1i(hatchStepUnif, step)

        # Prepare fill mask
        gl.glEnableVertexAttribArray(posAttrib)
        gl.glVertexAttribPointer(
            posAttrib, 2, gl.GL_FLOAT, gl.GL_FALSE, 0, self.vertices
        )

        gl.glEnable(gl.GL_STENCIL_TEST)
        gl.glStencilMask(1)
        gl.glStencilFunc(gl.GL_ALWAYS, 1, 1)
        gl.glStencilOp(gl.GL_INVERT, gl.GL_INVERT, gl.GL_INVERT)
        gl.glColorMask(gl.GL_FALSE, gl.GL_FALSE, gl.GL_FALSE, gl.GL_FALSE)
        gl.glDepthMask(gl.GL_FALSE)

        gl.glDrawElements(
            gl.GL_TRIANGLE_STRIP,
            len(self._indices),
            gl.GL_UNSIGNED_SHORT,
            self._indices,
        )

        gl.glStencilFunc(gl.GL_EQUAL, 1, 1)
        # Reset stencil while drawing
        gl.glStencilOp(gl.GL_ZERO, gl.GL_ZERO, gl.GL_ZERO)
        gl.glColorMask(gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE)
        gl.glDepthMask(gl.GL_TRUE)

        gl.glVertexAttribPointer(
            posAttrib, 2, gl.GL_FLOAT, gl.GL_FALSE, 0, self.bboxVertices
        )
        gl.glDrawArrays(gl.GL_TRIANGLE_STRIP, 0, len(self.bboxVertices))

        gl.glDisable(gl.GL_STENCIL_TEST)


# matrix ######################################################################


def mat4Ortho(left, right, bottom, top, near, far):
    """Orthographic projection matrix (row-major)"""
    return numpy.array(
        (
            (2.0 / (right - left), 0.0, 0.0, -(right + left) / float(right - left)),
            (0.0, 2.0 / (top - bottom), 0.0, -(top + bottom) / float(top - bottom)),
            (0.0, 0.0, -2.0 / (far - near), -(far + near) / float(far - near)),
            (0.0, 0.0, 0.0, 1.0),
        ),
        dtype=numpy.float64,
    )


def mat4Translate(x=0.0, y=0.0, z=0.0):
    """Translation matrix (row-major)"""
    return numpy.array(
        (
            (1.0, 0.0, 0.0, x),
            (0.0, 1.0, 0.0, y),
            (0.0, 0.0, 1.0, z),
            (0.0, 0.0, 0.0, 1.0),
        ),
        dtype=numpy.float64,
    )


def mat4Scale(sx=1.0, sy=1.0, sz=1.0):
    """Scale matrix (row-major)"""
    return numpy.array(
        (
            (sx, 0.0, 0.0, 0.0),
            (0.0, sy, 0.0, 0.0),
            (0.0, 0.0, sz, 0.0),
            (0.0, 0.0, 0.0, 1.0),
        ),
        dtype=numpy.float64,
    )


def mat4Identity():
    """Identity matrix"""
    return numpy.array(
        (
            (1.0, 0.0, 0.0, 0.0),
            (0.0, 1.0, 0.0, 0.0),
            (0.0, 0.0, 1.0, 0.0),
            (0.0, 0.0, 0.0, 1.0),
        ),
        dtype=numpy.float64,
    )



""" Fix the format of GitHub-generated release notes. """

import os.path
import re

root_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), ".."))

with open(os.path.join(root_dir, 'CHANGELOG.rst'), 'r') as f:
    content = f.read().split('/n')


with open(os.path.join(root_dir,'CHANGELOG_new.rst'), 'w') as f:
    for line in content:
        new_line = re.sub(r'by @.+ in https://github\.com/silx-kit/silx/pull/(\d+)', r'(PR #\1)', line)
        f.write(new_line)
        f.write('\n')



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""GUI to compare images"""

import sys
import logging
import argparse
import silx
from silx.gui import qt
from silx.app.utils import parseutils
from silx.app.compare.CompareImagesWindow import CompareImagesWindow

_logger = logging.getLogger(__name__)


file_description = """
Image data to compare (HDF5 file with path, EDF files, JPEG/PNG image files).
Data from HDF5 files can be accessed using dataset path and slicing as an URL: silx:../my_file.h5?path=/entry/data&slice=10
EDF file frames also can can be accessed using URL: fabio:../my_file.edf?slice=10
Using URL in command like usually have to be quoted: "URL".
"""


def createParser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("files", nargs=argparse.ZERO_OR_MORE, help=file_description)
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    parser.add_argument(
        "--use-opengl-plot",
        dest="use_opengl_plot",
        action="store_true",
        default=False,
        help="Use OpenGL for plots (instead of matplotlib)",
    )
    return parser


def mainQt(options):
    """Part of the main depending on Qt"""
    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    if options.use_opengl_plot:
        backend = "gl"
    else:
        backend = "mpl"

    settings = qt.QSettings(
        qt.QSettings.IniFormat, qt.QSettings.UserScope, "silx", "silx-compare", None
    )

    urls = list(parseutils.filenames_to_dataurls(options.files))

    if options.use_opengl_plot:
        # It have to be done after the settings (after the Viewer creation)
        silx.config.DEFAULT_PLOT_BACKEND = "opengl"

    app = qt.QApplication([])
    window = CompareImagesWindow(backend=backend, settings=settings)
    window.setAttribute(qt.Qt.WA_DeleteOnClose, True)

    # Note: Have to be before setUrls to have a proper resetZoom
    window.setVisible(True)

    window.setUrls(urls)

    app.exec()


def main(argv):
    parser = createParser()
    options = parser.parse_args(argv[1:])
    mainQt(options)


if __name__ == "__main__":
    main(sys.argv)



# /*##########################################################################
#
# Copyright (c) 2017-2024 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides the :class:`Scatter` item of the :class:`Plot`.
"""

__authors__ = ["T. Vincent", "P. Knobel"]
__license__ = "MIT"
__date__ = "29/03/2017"


from collections import namedtuple
import logging
import threading
import numpy
from matplotlib.tri import LinearTriInterpolator, Triangulation

from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, CancelledError

from ....utils.proxy import docstring
from ....math.combo import min_max
from ....math.histogram import Histogramnd
from ....utils.weakref import WeakList
from .core import PointsBase, ColormapMixIn, ScatterVisualizationMixIn
from .axis import Axis
from ._pick import PickingResult
from silx._utils import NP_OPTIONAL_COPY


_logger = logging.getLogger(__name__)


class _GreedyThreadPoolExecutor(ThreadPoolExecutor):
    """:class:`ThreadPoolExecutor` with an extra :meth:`submit_greedy` method."""

    def __init__(self, *args, **kwargs):
        super(_GreedyThreadPoolExecutor, self).__init__(*args, **kwargs)
        self.__futures = defaultdict(WeakList)
        self.__lock = threading.RLock()

    def submit_greedy(self, queue, fn, *args, **kwargs):
        """Same as :meth:`submit` but cancel previous tasks in given queue.

        This means that when a new task is submitted for a given queue,
        all other pending tasks of that queue are cancelled.

        :param queue: Identifier of the queue. This must be hashable.
        :param callable fn: The callable to call with provided extra arguments
        :return: Future corresponding to this task
        :rtype: concurrent.futures.Future
        """
        with self.__lock:
            # Cancel previous tasks in given queue
            for future in self.__futures.pop(queue, []):
                if not future.done():
                    future.cancel()

            future = super(_GreedyThreadPoolExecutor, self).submit(fn, *args, **kwargs)
            self.__futures[queue].append(future)

        return future


# Functions to guess grid shape from coordinates


def _get_z_line_length(array):
    """Return length of line if array is a Z-like 2D regular grid.

    :param numpy.ndarray array: The 1D array of coordinates to check
    :return: 0 if no line length could be found,
        else the number of element per line.
    :rtype: int
    """
    sign = numpy.sign(numpy.diff(array))
    if len(sign) == 0 or sign[0] == 0:  # We don't handle that
        return 0
    # Check this way to account for 0 sign (i.e., diff == 0)
    beginnings = numpy.where(sign == -sign[0])[0] + 1
    if len(beginnings) == 0:
        return 0
    length = beginnings[0]
    if numpy.all(numpy.equal(numpy.diff(beginnings), length)):
        return length
    return 0


def _guess_z_grid_shape(x, y):
    """Guess the shape of a grid from (x, y) coordinates.

    The grid might contain more elements than x and y,
    as the last line might be partly filled.

    :param numpy.ndarray x:
    :paran numpy.ndarray y:
    :returns: (order, (height, width)) of the regular grid,
        or None if could not guess one.
        'order' is 'row' if X (i.e., column) is the fast dimension, else 'column'.
    :rtype: Union[List(str,int),None]
    """
    width = _get_z_line_length(x)
    if width != 0:
        return "row", (int(numpy.ceil(len(x) / width)), width)
    else:
        height = _get_z_line_length(y)
        if height != 0:
            return "column", (height, int(numpy.ceil(len(y) / height)))
    return None


def is_monotonic(array):
    """Returns whether array is monotonic (increasing or decreasing).

    :param numpy.ndarray array: 1D array-like container.
    :returns: 1 if array is monotonically increasing,
       -1 if array is monotonically decreasing,
       0 if array is not monotonic
    :rtype: int
    """
    diff = numpy.diff(numpy.ravel(array))
    with numpy.errstate(invalid="ignore"):
        if numpy.all(diff >= 0):
            return 1
        elif numpy.all(diff <= 0):
            return -1
        else:
            return 0


def _guess_grid(x, y):
    """Guess a regular grid from the points.

    Result convention is (x, y)

    :param numpy.ndarray x: X coordinates of the points
    :param numpy.ndarray y: Y coordinates of the points
    :returns: (order, (height, width)
        order is 'row' or 'column'
    :rtype: Union[List[str,List[int]],None]
    """
    x, y = numpy.ravel(x), numpy.ravel(y)

    guess = _guess_z_grid_shape(x, y)
    if guess is not None:
        return guess

    else:
        # Cannot guess a regular grid
        # Let's assume it's a single line
        order = "row"  # or 'column' doesn't matter for a single line
        y_monotonic = is_monotonic(y)
        if is_monotonic(x) or y_monotonic:  # we can guess a line
            x_min, x_max = min_max(x)
            y_min, y_max = min_max(y)

            if not y_monotonic or x_max - x_min >= y_max - y_min:
                # x only is monotonic or both are and X varies more
                # line along X
                shape = 1, len(x)
            else:
                # y only is monotonic or both are and Y varies more
                # line along Y
                shape = len(y), 1

        else:  # Cannot guess a line from the points
            return None

    return order, shape


def _quadrilateral_grid_coords(points):
    """Compute an irregular grid of quadrilaterals from a set of points

    The input points are expected to lie on a grid.

    :param numpy.ndarray points:
       3D data set of 2D input coordinates (height, width, 2)
       height and width must be at least 2.
    :return: 3D dataset of 2D coordinates of the grid (height+1, width+1, 2)
    """
    assert points.ndim == 3
    assert points.shape[0] >= 2
    assert points.shape[1] >= 2
    assert points.shape[2] == 2

    dim0, dim1 = points.shape[:2]
    grid_points = numpy.zeros((dim0 + 1, dim1 + 1, 2), dtype=numpy.float64)

    # Compute inner points as mean of 4 neighbours
    neighbour_view = numpy.lib.stride_tricks.as_strided(
        points,
        shape=(dim0 - 1, dim1 - 1, 2, 2, points.shape[2]),
        strides=points.strides[:2] + points.strides[:2] + points.strides[-1:],
        writeable=False,
    )
    inner_points = numpy.mean(neighbour_view, axis=(2, 3))
    grid_points[1:-1, 1:-1] = inner_points

    # Compute 'vertical' sides
    # Alternative: grid_points[1:-1, [0, -1]] = points[:-1, [0, -1]] + points[1:, [0, -1]] - inner_points[:, [0, -1]]
    grid_points[1:-1, [0, -1], 0] = (
        points[:-1, [0, -1], 0] + points[1:, [0, -1], 0] - inner_points[:, [0, -1], 0]
    )
    grid_points[1:-1, [0, -1], 1] = inner_points[:, [0, -1], 1]

    # Compute 'horizontal' sides
    grid_points[[0, -1], 1:-1, 0] = inner_points[[0, -1], :, 0]
    grid_points[[0, -1], 1:-1, 1] = (
        points[[0, -1], :-1, 1] + points[[0, -1], 1:, 1] - inner_points[[0, -1], :, 1]
    )

    # Compute corners
    d0, d1 = [0, 0, -1, -1], [0, -1, -1, 0]
    grid_points[d0, d1] = 2 * points[d0, d1] - inner_points[d0, d1]
    return grid_points


def _quadrilateral_grid_as_triangles(points):
    """Returns the points and indices to make a grid of quadrilaterals

    :param numpy.ndarray points:
        3D array of points (height, width, 2)
    :return: triangle corners (4 * N, 2), triangle indices (2 * N, 3)
        With N = height * width, the number of input points
    """
    nbpoints = numpy.prod(points.shape[:2])

    grid = _quadrilateral_grid_coords(points)
    coords = numpy.empty((4 * nbpoints, 2), dtype=grid.dtype)
    coords[::4] = grid[:-1, :-1].reshape(-1, 2)
    coords[1::4] = grid[1:, :-1].reshape(-1, 2)
    coords[2::4] = grid[:-1, 1:].reshape(-1, 2)
    coords[3::4] = grid[1:, 1:].reshape(-1, 2)

    indices = numpy.empty((2 * nbpoints, 3), dtype=numpy.uint32)
    indices[::2, 0] = numpy.arange(0, 4 * nbpoints, 4)
    indices[::2, 1] = numpy.arange(1, 4 * nbpoints, 4)
    indices[::2, 2] = numpy.arange(2, 4 * nbpoints, 4)
    indices[1::2, 0] = indices[::2, 1]
    indices[1::2, 1] = indices[::2, 2]
    indices[1::2, 2] = numpy.arange(3, 4 * nbpoints, 4)

    return coords, indices


_RegularGridInfo = namedtuple(
    "_RegularGridInfo", ["bounds", "origin", "scale", "shape", "order"]
)


_HistogramInfo = namedtuple(
    "_HistogramInfo", ["mean", "count", "sum", "origin", "scale", "shape"]
)


class Scatter(PointsBase, ColormapMixIn, ScatterVisualizationMixIn):
    """Description of a scatter"""

    _DEFAULT_SELECTABLE = True
    """Default selectable state for scatter plots"""

    _SUPPORTED_SCATTER_VISUALIZATION = (
        ScatterVisualizationMixIn.Visualization.POINTS,
        ScatterVisualizationMixIn.Visualization.SOLID,
        ScatterVisualizationMixIn.Visualization.REGULAR_GRID,
        ScatterVisualizationMixIn.Visualization.IRREGULAR_GRID,
        ScatterVisualizationMixIn.Visualization.BINNED_STATISTIC,
    )
    """Overrides supported Visualizations"""

    def __init__(self):
        PointsBase.__init__(self)
        ColormapMixIn.__init__(self)
        ScatterVisualizationMixIn.__init__(self)
        self._value = ()
        self.__alpha = None
        # Cache Delaunay triangulation future object
        self.__triangulationFuture = None
        # Cache interpolator future object
        self.__interpolatorFuture = None
        self.__executor = None

        # Cache triangles: x, y, indices
        self.__cacheTriangles = None, None, None

        # Cache regular grid and histogram info
        self.__cacheRegularGridInfo = None
        self.__cacheHistogramInfo = None

    def _updateColormappedData(self):
        """Update the colormapped data, to be called when changed"""
        if self.getVisualization() is self.Visualization.BINNED_STATISTIC:
            histoInfo = self.__getHistogramInfo()
            if histoInfo is None:
                data = None
            else:
                data = getattr(
                    histoInfo,
                    self.getVisualizationParameter(
                        self.VisualizationParameter.BINNED_STATISTIC_FUNCTION
                    ),
                )
        else:
            data = self.getValueData(copy=False)
        self._setColormappedData(data, copy=False)

    @docstring(ScatterVisualizationMixIn)
    def setVisualization(self, mode):
        previous = self.getVisualization()
        if super().setVisualization(mode):
            if bool(mode is self.Visualization.BINNED_STATISTIC) ^ bool(
                previous is self.Visualization.BINNED_STATISTIC
            ):
                self._updateColormappedData()
            return True
        else:
            return False

    @docstring(ScatterVisualizationMixIn)
    def setVisualizationParameter(self, parameter, value):
        parameter = self.VisualizationParameter(parameter)

        if super(Scatter, self).setVisualizationParameter(parameter, value):
            if parameter in (
                self.VisualizationParameter.GRID_BOUNDS,
                self.VisualizationParameter.GRID_MAJOR_ORDER,
                self.VisualizationParameter.GRID_SHAPE,
            ):
                self.__cacheRegularGridInfo = None

            if parameter in (
                self.VisualizationParameter.BINNED_STATISTIC_SHAPE,
                self.VisualizationParameter.BINNED_STATISTIC_FUNCTION,
                self.VisualizationParameter.DATA_BOUNDS_HINT,
            ):
                if parameter in (
                    self.VisualizationParameter.BINNED_STATISTIC_SHAPE,
                    self.VisualizationParameter.DATA_BOUNDS_HINT,
                ):
                    self.__cacheHistogramInfo = None  # Clean-up cache
                if self.getVisualization() is self.Visualization.BINNED_STATISTIC:
                    self._updateColormappedData()
            return True
        else:
            return False

    @docstring(ScatterVisualizationMixIn)
    def getCurrentVisualizationParameter(self, parameter):
        value = self.getVisualizationParameter(parameter)
        if (
            parameter is self.VisualizationParameter.DATA_BOUNDS_HINT
            or value is not None
        ):
            return value  # Value has been set, return it

        elif parameter is self.VisualizationParameter.GRID_BOUNDS:
            grid = self.__getRegularGridInfo()
            return None if grid is None else grid.bounds

        elif parameter is self.VisualizationParameter.GRID_MAJOR_ORDER:
            grid = self.__getRegularGridInfo()
            return None if grid is None else grid.order

        elif parameter is self.VisualizationParameter.GRID_SHAPE:
            grid = self.__getRegularGridInfo()
            return None if grid is None else grid.shape

        elif parameter is self.VisualizationParameter.BINNED_STATISTIC_SHAPE:
            info = self.__getHistogramInfo()
            return None if info is None else info.shape

        else:
            raise NotImplementedError()

    def __getRegularGridInfo(self):
        """Get grid info"""
        if self.__cacheRegularGridInfo is None:
            shape = self.getVisualizationParameter(
                self.VisualizationParameter.GRID_SHAPE
            )
            order = self.getVisualizationParameter(
                self.VisualizationParameter.GRID_MAJOR_ORDER
            )
            if shape is None or order is None:
                guess = _guess_grid(
                    self.getXData(copy=False), self.getYData(copy=False)
                )
                if guess is None:
                    _logger.warning(
                        "Cannot guess a grid: Cannot display as regular grid image"
                    )
                    return None
                if shape is None:
                    shape = guess[1]
                if order is None:
                    order = guess[0]

            nbpoints = len(self.getXData(copy=False))
            if nbpoints > shape[0] * shape[1]:
                # More data points that provided grid shape: enlarge grid
                _logger.warning(
                    "More data points than provided grid shape size: extends grid"
                )
                dim0, dim1 = shape
                if order == "row":  # keep dim1, enlarge dim0
                    dim0 = nbpoints // dim1 + (1 if nbpoints % dim1 else 0)
                else:  # keep dim0, enlarge dim1
                    dim1 = nbpoints // dim0 + (1 if nbpoints % dim0 else 0)
                shape = dim0, dim1

            bounds = self.getVisualizationParameter(
                self.VisualizationParameter.GRID_BOUNDS
            )
            if bounds is None:
                x, y = self.getXData(copy=False), self.getYData(copy=False)
                min_, max_ = min_max(x)
                xRange = (min_, max_) if (x[0] - min_) < (max_ - x[0]) else (max_, min_)
                min_, max_ = min_max(y)
                yRange = (min_, max_) if (y[0] - min_) < (max_ - y[0]) else (max_, min_)
                bounds = (xRange[0], yRange[0]), (xRange[1], yRange[1])

            begin, end = bounds
            scale = (
                (end[0] - begin[0]) / max(1, shape[1] - 1),
                (end[1] - begin[1]) / max(1, shape[0] - 1),
            )
            if scale[0] == 0 and scale[1] == 0:
                scale = 1.0, 1.0
            elif scale[0] == 0:
                scale = scale[1], scale[1]
            elif scale[1] == 0:
                scale = scale[0], scale[0]

            origin = begin[0] - 0.5 * scale[0], begin[1] - 0.5 * scale[1]

            self.__cacheRegularGridInfo = _RegularGridInfo(
                bounds=bounds, origin=origin, scale=scale, shape=shape, order=order
            )

        return self.__cacheRegularGridInfo

    def __getHistogramInfo(self):
        """Get histogram info"""
        if self.__cacheHistogramInfo is None:
            shape = self.getVisualizationParameter(
                self.VisualizationParameter.BINNED_STATISTIC_SHAPE
            )
            if shape is None:
                shape = 100, 100  # TODO compute auto shape

            x, y, values = self.getData(copy=False)[:3]
            if len(x) == 0:  # No histogram
                return None

            if not numpy.issubdtype(x.dtype, numpy.floating):
                x = x.astype(numpy.float64)
            if not numpy.issubdtype(y.dtype, numpy.floating):
                y = y.astype(numpy.float64)
            if not numpy.issubdtype(values.dtype, numpy.floating):
                values = values.astype(numpy.float64)

            ranges = (tuple(min_max(y, finite=True)), tuple(min_max(x, finite=True)))
            rangesHint = self.getVisualizationParameter(
                self.VisualizationParameter.DATA_BOUNDS_HINT
            )
            if rangesHint is not None:
                ranges = tuple(
                    (min(dataMin, hintMin), max(dataMax, hintMax))
                    for (dataMin, dataMax), (hintMin, hintMax) in zip(
                        ranges, rangesHint
                    )
                )

            points = numpy.transpose(numpy.array((y, x)))
            counts, sums, bin_edges = Histogramnd(
                points, histo_range=ranges, n_bins=shape, weights=values
            )
            yEdges, xEdges = bin_edges
            origin = xEdges[0], yEdges[0]
            scale = (
                (xEdges[-1] - xEdges[0]) / (len(xEdges) - 1),
                (yEdges[-1] - yEdges[0]) / (len(yEdges) - 1),
            )

            with numpy.errstate(divide="ignore", invalid="ignore"):
                histo = sums / counts

            self.__cacheHistogramInfo = _HistogramInfo(
                mean=histo,
                count=counts,
                sum=sums,
                origin=origin,
                scale=scale,
                shape=shape,
            )

        return self.__cacheHistogramInfo

    def __applyColormapToData(self):
        """Compute colors by applying colormap to values.

        :returns: Array of RGBA colors
        """
        cmap = self.getColormap()
        rgbacolors = cmap.applyToData(self)

        if self.__alpha is not None:
            rgbacolors[:, -1] = (rgbacolors[:, -1] * self.__alpha).astype(numpy.uint8)
        return rgbacolors

    def _addBackendRenderer(self, backend):
        """Update backend renderer"""
        # Filter-out values <= 0
        xFiltered, yFiltered, valueFiltered, xerror, yerror = self.getData(
            copy=False, displayed=True
        )

        # Remove not finite numbers (this includes filtered out x, y <= 0)
        mask = numpy.logical_and(numpy.isfinite(xFiltered), numpy.isfinite(yFiltered))
        xFiltered = xFiltered[mask]
        yFiltered = yFiltered[mask]

        if len(xFiltered) == 0:
            return None  # No data to display, do not add renderer to backend

        visualization = self.getVisualization()

        if visualization is self.Visualization.BINNED_STATISTIC:
            plot = self.getPlot()
            if (
                plot is None
                or plot.getXAxis().getScale() != Axis.LINEAR
                or plot.getYAxis().getScale() != Axis.LINEAR
            ):
                # Those visualizations are not available with log scaled axes
                return None

            histoInfo = self.__getHistogramInfo()
            if histoInfo is None:
                return None
            data = getattr(
                histoInfo,
                self.getVisualizationParameter(
                    self.VisualizationParameter.BINNED_STATISTIC_FUNCTION
                ),
            )

            return backend.addImage(
                data=data,
                origin=histoInfo.origin,
                scale=histoInfo.scale,
                colormap=self.getColormap(),
                alpha=self.getAlpha(),
            )

        elif visualization is self.Visualization.POINTS:
            rgbacolors = self.__applyColormapToData()
            return backend.addCurve(
                xFiltered,
                yFiltered,
                color=rgbacolors[mask],
                gapcolor=None,
                symbol=self.getSymbol(),
                linewidth=0,
                linestyle="",
                yaxis="left",
                xerror=xerror,
                yerror=yerror,
                fill=False,
                alpha=self.getAlpha(),
                symbolsize=self.getSymbolSize(),
                baseline=None,
            )

        else:
            plot = self.getPlot()
            if (
                plot is None
                or plot.getXAxis().getScale() != Axis.LINEAR
                or plot.getYAxis().getScale() != Axis.LINEAR
            ):
                # Those visualizations are not available with log scaled axes
                return None

            if visualization is self.Visualization.SOLID:
                try:
                    triangulation = self._getTriangulationFuture().result()
                except (RuntimeError, ValueError):
                    _logger.warning(
                        "Cannot get a triangulation: Cannot display as solid surface"
                    )
                    return None
                else:
                    rgbacolors = self.__applyColormapToData()
                    triangles = triangulation.triangles.astype(numpy.int32)
                    return backend.addTriangles(
                        xFiltered,
                        yFiltered,
                        triangles,
                        color=rgbacolors[mask],
                        alpha=self.getAlpha(),
                    )

            elif visualization is self.Visualization.REGULAR_GRID:
                gridInfo = self.__getRegularGridInfo()
                if gridInfo is None:
                    return None

                dim0, dim1 = gridInfo.shape
                if gridInfo.order == "column":  # transposition needed
                    dim0, dim1 = dim1, dim0

                values = self.getValueData(copy=False)
                if self.__alpha is None and len(values) == dim0 * dim1:
                    image = values.reshape(dim0, dim1)
                else:
                    # The points do not fill the whole image
                    if self.__alpha is None and numpy.issubdtype(
                        values.dtype, numpy.floating
                    ):
                        image = numpy.empty(dim0 * dim1, dtype=values.dtype)
                        image[: len(values)] = values
                        image[len(values) :] = float("nan")  # Transparent pixels
                        image.shape = dim0, dim1
                    else:  # Per value alpha or no NaN, so convert to RGBA
                        rgbacolors = self.__applyColormapToData()
                        image = numpy.empty((dim0 * dim1, 4), dtype=numpy.uint8)
                        image[: len(rgbacolors)] = rgbacolors
                        image[len(rgbacolors) :] = (0, 0, 0, 0)  # Transparent pixels
                        image.shape = dim0, dim1, 4

                if gridInfo.order == "column":
                    if image.ndim == 2:
                        image = numpy.transpose(image)
                    else:
                        image = numpy.transpose(image, axes=(1, 0, 2))

                if image.ndim == 2:
                    colormap = self.getColormap()
                    if colormap.isAutoscale():
                        # Avoid backend to compute autoscale: use item cache
                        colormap = colormap.copy()
                        colormap.setVRange(*colormap.getColormapRange(self))
                else:
                    colormap = None

                return backend.addImage(
                    data=image,
                    origin=gridInfo.origin,
                    scale=gridInfo.scale,
                    colormap=colormap,
                    alpha=self.getAlpha(),
                )

            elif visualization is self.Visualization.IRREGULAR_GRID:
                gridInfo = self.__getRegularGridInfo()
                if gridInfo is None:
                    return None

                shape = gridInfo.shape
                if shape is None:  # No shape, no display
                    return None

                rgbacolors = self.__applyColormapToData()

                nbpoints = len(xFiltered)
                if nbpoints == 1:
                    # single point, render as a square points
                    return backend.addCurve(
                        xFiltered,
                        yFiltered,
                        color=rgbacolors[mask],
                        gapcolor=None,
                        symbol="s",
                        linewidth=0,
                        linestyle="",
                        yaxis="left",
                        xerror=None,
                        yerror=None,
                        fill=False,
                        alpha=self.getAlpha(),
                        symbolsize=7,
                        baseline=None,
                    )

                # Make shape include all points
                gridOrder = gridInfo.order
                if nbpoints != numpy.prod(shape):
                    if gridOrder == "row":
                        shape = int(numpy.ceil(nbpoints / shape[1])), shape[1]
                    else:  # column-major order
                        shape = shape[0], int(numpy.ceil(nbpoints / shape[0]))

                if shape[0] < 2 or shape[1] < 2:  # Single line, at least 2 points
                    points = numpy.zeros((2, nbpoints, 3), dtype=numpy.float64)
                    # Use row/column major depending on shape, not on info value
                    gridOrder = "row" if shape[0] == 1 else "column"

                    if gridOrder == "row":
                        points[0, :, 0] = xFiltered
                        points[0, :, 1] = yFiltered
                    else:  # column-major order
                        points[0, :, 0] = yFiltered
                        points[0, :, 1] = xFiltered

                    # Add a second line that will be clipped in the end
                    points[1, :-1] = (
                        points[0, :-1]
                        + numpy.cross(points[0, 1:] - points[0, :-1], (0.0, 0.0, 1.0))
                    )
                    points[1, -1] = (
                        points[0, -1]
                        + numpy.cross(points[0, -1] - points[0, -2], (0.0, 0.0, 1.0))
                    )

                    points = points[:, :, :2]

                elif gridOrder == "row":  # row-major order
                    if nbpoints != numpy.prod(shape):
                        points = numpy.empty(
                            (numpy.prod(shape), 2), dtype=numpy.float64
                        )
                        points[:nbpoints, 0] = xFiltered
                        points[:nbpoints, 1] = yFiltered
                        # Index of last element of last fully filled row
                        index = (nbpoints // shape[1]) * shape[1]
                        points[nbpoints:, 0] = xFiltered[
                            index - (numpy.prod(shape) - nbpoints) : index
                        ]
                        points[nbpoints:, 1] = yFiltered[-1]
                    else:
                        points = numpy.transpose((xFiltered, yFiltered))
                    points.shape = shape[0], shape[1], 2

                else:  # column-major order
                    if nbpoints != numpy.prod(shape):
                        points = numpy.empty(
                            (numpy.prod(shape), 2), dtype=numpy.float64
                        )
                        points[:nbpoints, 0] = yFiltered
                        points[:nbpoints, 1] = xFiltered
                        # Index of last element of last fully filled column
                        index = (nbpoints // shape[0]) * shape[0]
                        points[nbpoints:, 0] = yFiltered[
                            index - (numpy.prod(shape) - nbpoints) : index
                        ]
                        points[nbpoints:, 1] = xFiltered[-1]
                    else:
                        points = numpy.transpose((yFiltered, xFiltered))
                    points.shape = shape[1], shape[0], 2

                coords, indices = _quadrilateral_grid_as_triangles(points)

                # Remove unused extra triangles
                coords = coords[: 4 * nbpoints]
                indices = indices[: 2 * nbpoints]

                if gridOrder == "row":
                    x, y = coords[:, 0], coords[:, 1]
                else:  # column-major order
                    y, x = coords[:, 0], coords[:, 1]

                rgbacolors = rgbacolors[mask]  # Filter-out not finite points
                gridcolors = numpy.empty(
                    (4 * nbpoints, rgbacolors.shape[-1]), dtype=rgbacolors.dtype
                )
                for first in range(4):
                    gridcolors[first::4] = rgbacolors[:nbpoints]

                return backend.addTriangles(
                    x, y, indices, color=gridcolors, alpha=self.getAlpha()
                )

            else:
                _logger.error("Unhandled visualization %s", visualization)
                return None

    @docstring(PointsBase)
    def pick(self, x, y):
        result = super(Scatter, self).pick(x, y)

        if result is not None:
            visualization = self.getVisualization()

            if visualization is self.Visualization.IRREGULAR_GRID:
                # Specific handling of picking for the irregular grid mode
                index = result.getIndices(copy=False)[0] // 4
                result = PickingResult(self, (index,))

            elif visualization is self.Visualization.REGULAR_GRID:
                # Specific handling of picking for the regular grid mode
                picked = result.getIndices(copy=False)
                if picked is None:
                    return None
                row, column = picked[0][0], picked[1][0]

                gridInfo = self.__getRegularGridInfo()
                if gridInfo is None:
                    return None

                if gridInfo.order == "row":
                    index = row * gridInfo.shape[1] + column
                else:
                    index = row + column * gridInfo.shape[0]
                if index >= len(
                    self.getXData(copy=False)
                ):  # OK as long as not log scale
                    return None  # Image can be larger than scatter

                result = PickingResult(self, (index,))

            elif visualization is self.Visualization.BINNED_STATISTIC:
                picked = result.getIndices(copy=False)
                if picked is None or len(picked) == 0 or len(picked[0]) == 0:
                    return None
                row, col = picked[0][0], picked[1][0]
                histoInfo = self.__getHistogramInfo()
                if histoInfo is None:
                    return None
                sx, sy = histoInfo.scale
                ox, oy = histoInfo.origin
                xdata = self.getXData(copy=False)
                ydata = self.getYData(copy=False)
                indices = numpy.nonzero(
                    numpy.logical_and(
                        numpy.logical_and(
                            xdata >= ox + sx * col, xdata < ox + sx * (col + 1)
                        ),
                        numpy.logical_and(
                            ydata >= oy + sy * row, ydata < oy + sy * (row + 1)
                        ),
                    )
                )[0]
                result = None if len(indices) == 0 else PickingResult(self, indices)

        return result

    def __getExecutor(self):
        """Returns async greedy executor

        :rtype: _GreedyThreadPoolExecutor
        """
        if self.__executor is None:
            self.__executor = _GreedyThreadPoolExecutor(max_workers=2)
        return self.__executor

    def _getTriangulationFuture(self):
        """Returns a :class:`Future` which result is the Triangulation object.

        :rtype: concurrent.futures.Future
        """
        if self.__triangulationFuture is None or self.__triangulationFuture.cancelled():
            # Need to init a new delaunay
            x, y = self.getData(copy=False)[:2]
            # Remove not finite points
            mask = numpy.logical_and(numpy.isfinite(x), numpy.isfinite(y))

            self.__triangulationFuture = self.__getExecutor().submit_greedy(
                "Triangulation", Triangulation, x[mask], y[mask]
            )

        return self.__triangulationFuture

    @staticmethod
    def __initInterpolator(triangulationFuture, values):
        """Returns an interpolator for the given data points

        :param concurrent.futures.Future triangulationFuture:
            Future object which result is a Triangulation object
        :param numpy.ndarray values: The data value of valid points.
        :rtype: Union[callable,None]
        """
        # Wait for Triangulation to complete
        try:
            triangulation = triangulationFuture.result()
        except (RuntimeError, ValueError):
            return None  # triangulation failed
        except CancelledError:
            return None

        return LinearTriInterpolator(triangulation, values)

    def _getInterpolatorFuture(self):
        """Returns a :class:`Future` which result is the interpolator.

        The interpolator is a callable taking an array Nx2 of points
        as a single argument.
        The :class:`Future` result is None in case the interpolator cannot
        be initialized.

        :rtype: concurrent.futures.Future
        """
        if self.__interpolatorFuture is None or self.__interpolatorFuture.cancelled():
            # Need to init a new interpolator
            x, y, values = self.getData(copy=False)[:3]
            # Remove not finite points
            mask = numpy.logical_and(numpy.isfinite(x), numpy.isfinite(y))
            x, y, values = x[mask], y[mask], values[mask]

            self.__interpolatorFuture = self.__getExecutor().submit_greedy(
                "interpolator",
                self.__initInterpolator,
                self._getTriangulationFuture(),
                values,
            )
        return self.__interpolatorFuture

    def _filterData(self, xPositive: bool, yPositive: bool):
        """Filter out errors<0 and values with x or y <= 0 on log axes

        :param bool xPositive: True to filter arrays according to X coords.
        :param bool yPositive: True to filter arrays according to Y coords.
        :return: The filtered arrays or unchanged object if not filtering needed
        :rtype: (x, y, value, xerror, yerror)
        """
        # overloaded from PointsBase to filter also value.
        value = self.getValueData(copy=False)

        if xPositive or yPositive:
            clipped = self._getClippingBoolArray(xPositive, yPositive)

            if numpy.any(clipped):
                # copy to keep original array and convert to float
                value = numpy.array(value, copy=True, dtype=numpy.float64)
                value[clipped] = numpy.nan

        x, y, xerror, yerror = PointsBase._filterData(self, xPositive, yPositive)

        return x, y, value, xerror, yerror

    def getValueData(self, copy=True):
        """Returns the value assigned to the scatter data points.

        :param copy: True (Default) to get a copy,
                     False to use internal representation (do not modify!)
        :rtype: numpy.ndarray
        """
        return numpy.array(self._value, copy=copy or NP_OPTIONAL_COPY)

    def getAlphaData(self, copy=True):
        """Returns the alpha (transparency) assigned to the scatter data points.

        :param copy: True (Default) to get a copy,
                     False to use internal representation (do not modify!)
        :rtype: numpy.ndarray
        """
        return numpy.array(self.__alpha, copy=copy or NP_OPTIONAL_COPY)

    def getData(self, copy=True, displayed=False):
        """Returns the x, y coordinates and the value of the data points

        :param copy: True (Default) to get a copy,
                     False to use internal representation (do not modify!)
        :param bool displayed: True to only get curve points that are displayed
                               in the plot. Default: False.
                               Note: If plot has log scale, negative points
                               are not displayed.
        :returns: (x, y, value, xerror, yerror)
        :rtype: 5-tuple of numpy.ndarray
        """
        if displayed:
            data = self._getCachedData()
            if data is not None:
                assert len(data) == 5
                return data

        return (
            self.getXData(copy),
            self.getYData(copy),
            self.getValueData(copy),
            self.getXErrorData(copy),
            self.getYErrorData(copy),
        )

    # reimplemented from PointsBase to handle `value`
    def setData(self, x, y, value, xerror=None, yerror=None, alpha=None, copy=True):
        """Set the data of the scatter.

        :param numpy.ndarray x: The data corresponding to the x coordinates.
        :param numpy.ndarray y: The data corresponding to the y coordinates.
        :param numpy.ndarray value: The data corresponding to the value of
                                    the data points.
        :param xerror: Values with the uncertainties on the x values
        :type xerror: A float, or a numpy.ndarray of float32.
                      If it is an array, it can either be a 1D array of
                      same length as the data or a 2D array with 2 rows
                      of same length as the data: row 0 for lower errors,
                      row 1 for upper errors.
        :param yerror: Values with the uncertainties on the y values
        :type yerror: A float, or a numpy.ndarray of float32. See xerror.
        :param alpha: Values with the transparency (between 0 and 1)
        :type alpha: A float, or a numpy.ndarray of float32
        :param bool copy: True make a copy of the data (default),
                          False to use provided arrays.
        """
        value = numpy.array(value, copy=copy or NP_OPTIONAL_COPY)
        assert value.ndim == 1
        assert len(x) == len(value)

        # Convert complex data
        if numpy.iscomplexobj(value):
            _logger.warning("Converting value data to absolute value to plot it.")
            value = numpy.absolute(value)

        # Reset triangulation and interpolator
        if self.__triangulationFuture is not None:
            self.__triangulationFuture.cancel()
            self.__triangulationFuture = None
        if self.__interpolatorFuture is not None:
            self.__interpolatorFuture.cancel()
            self.__interpolatorFuture = None

        # Data changed, this needs update
        self.__cacheRegularGridInfo = None
        self.__cacheHistogramInfo = None

        self._value = value

        if alpha is not None:
            # Make sure alpha is an array of float in [0, 1]
            alpha = numpy.array(alpha, copy=copy or NP_OPTIONAL_COPY)
            assert alpha.ndim == 1
            assert len(x) == len(alpha)
            if alpha.dtype.kind != "f":
                alpha = alpha.astype(numpy.float32)
            if numpy.any(numpy.logical_or(alpha < 0.0, alpha > 1.0)):
                alpha = numpy.clip(alpha, 0.0, 1.0)
        self.__alpha = alpha

        # set x, y, xerror, yerror

        # call self._updated + plot._invalidateDataRange()
        PointsBase.setData(self, x, y, xerror, yerror, copy)

        self._updateColormappedData()



# /*##########################################################################
#
# Copyright (c) 2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Test PlotWidget features related to data margins"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "11/05/2023"

import numpy
import pytest


def testDefaultDataMargins(plotWidget):
    """Test default PlotWidget data margins: No margins"""
    assert plotWidget.getDataMargins() == (0, 0, 0, 0)


def testResetZoomDataMarginsLinearAxes(qapp, plotWidget):
    """Test PlotWidget.setDataMargins effect on resetZoom with linear axis scales"""

    margins = 0.1, 0.2, 0.3, 0.4
    plotWidget.setDataMargins(*margins)

    plotWidget.resetZoom()
    qapp.processEvents()

    retrievedMargins = plotWidget.getDataMargins()
    assert retrievedMargins == margins

    dataRange = 100 - 1
    expectedXLimits = 1 - 0.1 * dataRange, 100 + 0.2 * dataRange
    expectedYLimits = 1 - 0.3 * dataRange, 100 + 0.4 * dataRange

    assert plotWidget.getXAxis().getLimits() == expectedXLimits
    assert plotWidget.getYAxis().getLimits() == expectedYLimits
    assert plotWidget.getYAxis(axis="right").getLimits() == expectedYLimits


def testResetZoomDataMarginsLogAxes(qapp, plotWidget):
    """Test PlotWidget.setDataMargins effect on resetZoom with log axis scales"""
    plotWidget.getXAxis().setScale("log")
    plotWidget.getYAxis().setScale("log")

    dataMargins = 0.1, 0.2, 0.3, 0.4
    plotWidget.setDataMargins(*dataMargins)

    plotWidget.resetZoom()
    qapp.processEvents()

    retrievedMargins = plotWidget.getDataMargins()
    assert retrievedMargins == dataMargins

    logMin, logMax = numpy.log10(1), numpy.log10(100)
    logRange = logMax - logMin
    expectedXLimits = pow(10.0, logMin - 0.1 * logRange), pow(
        10.0, logMax + 0.2 * logRange
    )
    expectedYLimits = pow(10.0, logMin - 0.3 * logRange), pow(
        10.0, logMax + 0.4 * logRange
    )

    assert plotWidget.getXAxis().getLimits() == expectedXLimits
    assert plotWidget.getYAxis().getLimits() == expectedYLimits
    assert plotWidget.getYAxis(axis="right").getLimits() == expectedYLimits


@pytest.mark.parametrize("margins", [False, True, (0, 0, 0, 0)])
def testSetLimitsNoDataMargins(plotWidget, margins):
    """Test PlotWidget.setLimits without data margins"""
    xlimits = 1, 2
    ylimits = 3, 4
    y2limits = 5, 6
    plotWidget.setLimits(*xlimits, *ylimits, *y2limits, margins=margins)

    assert plotWidget.getXAxis().getLimits() == xlimits
    assert plotWidget.getYAxis().getLimits() == ylimits
    assert plotWidget.getYAxis(axis="right").getLimits() == y2limits


@pytest.mark.parametrize(
    "margins,expectedLimits",
    [
        # margins=False: use limits as is
        (
            False,
            (1, 2, 3, 4, 5, 6),
        ),
        # margins=True: apply data margins
        (
            True,
            (1 - 0.1, 2 + 0.2, 3 - 0.3, 4 + 0.4, 5 - 0.3, 6 + 0.4),
        ),
        # margins=tuple: apply provided margins
        (
            (0.4, 0.3, 0.2, 0.1),
            (1 - 0.4, 2 + 0.3, 3 - 0.2, 4 + 0.1, 5 - 0.2, 6 + 0.1),
        ),
    ],
)
def testSetLimitsWithDataMargins(qapp, plotWidget, margins, expectedLimits):
    """Test PlotWidget.setLimits with data margins"""
    dataMargins = 0.1, 0.2, 0.3, 0.4
    limits = 1, 2, 3, 4, 5, 6

    plotWidget.setDataMargins(*dataMargins)
    plotWidget.setLimits(*limits, margins=margins)
    qapp.processEvents()

    retrievedLimits = (
        *plotWidget.getXAxis().getLimits(),
        *plotWidget.getYAxis().getLimits(),
        *plotWidget.getYAxis(axis="right").getLimits(),
    )
    assert retrievedLimits == expectedLimits



# /*##########################################################################
#
# Copyright (c) 2004-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Module contains a float editor
"""
from __future__ import annotations


__authors__ = ["V.A. Sole", "T. Vincent"]
__license__ = "MIT"
__date__ = "02/10/2017"

from .. import qt


class FloatEdit(qt.QLineEdit):
    """Field to edit a float value.

    The value can be modified with :meth:`value` and :meth:`setValue`.

    The property :meth:`widgetResizable` allow to change the default
    behaviour in order to automatically resize the widget to the displayed value.
    Use :meth:`setMinimumWidth` to enforce the minimum width.

    :param parent: Parent of the widget
    :param value: The value to set the QLineEdit to.
    """

    _QLineEditPrivateHorizontalMargin = 2
    """Constant from Qt source code"""

    def __init__(self, parent: qt.QWidget | None = None, value: float | None = None):
        qt.QLineEdit.__init__(self, parent)
        validator = qt.QDoubleValidator(self)
        self.__widgetResizable: bool = False
        self.__minimumWidth = 30
        """Store the minimum width requested by the user, the real one is
        dynamic"""
        self.setValidator(validator)
        self.setAlignment(qt.Qt.AlignRight)
        self.textChanged.connect(self.__textChanged)
        if value is not None:
            self.setValue(value)

    def value(self) -> float:
        """Return the QLineEdit current value as a float."""
        text = self.text()
        value, validated = self.validator().locale().toDouble(text)
        if not validated:
            self.setValue(value)
        return value

    def setValue(self, value: float):
        """Set the current value of the LineEdit

        :param value: The value to set the QLineEdit to.
        """
        locale = self.validator().locale()
        if qt.BINDING == "PySide6":
            # Fix for PySide6 not selecting the right method
            text = locale.toString(float(value), "g")
        else:
            text = locale.toString(float(value))

        self.setText(text)
        if self.__widgetResizable:
            self.__forceMinimumWidthFromContent()

    def __textChanged(self, text: str):
        if self.__widgetResizable:
            self.__forceMinimumWidthFromContent()

    def widgetResizable(self) -> bool:
        """
        Returns whether or not the widget auto resizes itself based on it's content
        """
        return self.__widgetResizable

    def setWidgetResizable(self, resizable: bool):
        """
        If true, the widget will automatically resize itself to its displayed content.

        This avoids to have to scroll to see the widget's content, and allow to take
        advantage of extra space.
        """
        if self.__widgetResizable == resizable:
            return
        self.__widgetResizable = resizable
        self.updateGeometry()
        if resizable:
            self.__forceMinimumWidthFromContent()
        else:
            qt.QLineEdit.setMinimumWidth(self, self.__minimumWidth)

    def __minimumWidthFromContent(self) -> int:
        """Minimum size for the widget to properly read the actual number"""
        text = self.text()
        font = self.font()
        metrics = qt.QFontMetrics(font)
        margins = self.textMargins()
        width = (
            metrics.horizontalAdvance(text)
            + self._QLineEditPrivateHorizontalMargin * 2
            + margins.left()
            + margins.right()
        )
        width = max(self.__minimumWidth, width)
        opt = qt.QStyleOptionFrame()
        self.initStyleOption(opt)
        s = self.style().sizeFromContents(
            qt.QStyle.CT_LineEdit, opt, qt.QSize(width, self.height())
        )
        return s.width()

    def sizeHint(self) -> qt.QSize:
        sizeHint = qt.QLineEdit.sizeHint(self)
        if not self.__widgetResizable:
            return sizeHint
        width = self.__minimumWidthFromContent()
        return qt.QSize(width, sizeHint.height())

    def __forceMinimumWidthFromContent(self):
        width = self.__minimumWidthFromContent()
        qt.QLineEdit.setMinimumWidth(self, width)
        self.updateGeometry()

    def setMinimumWidth(self, width: int):
        self.__minimumWidth = width
        qt.QLineEdit.setMinimumWidth(self, width)
        self.updateGeometry()

    def minimumWidth(self) -> int:
        """Returns the user defined minimum width."""
        return self.__minimumWidth



from silx.utils.deprecation import deprecated_warning

deprecated_warning(
    "Module",
    "silx.third_party.TiffIO",
    since_version="2.0.0",
    replacement="fabio.TiffIO",
)

from fabio.TiffIO import *



# /*##########################################################################
#
# Copyright (c) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tests for FloatEdit"""

__license__ = "MIT"

import pytest
from silx.gui import qt
from silx.gui.widgets.FloatEdit import FloatEdit


@pytest.fixture
def floatEdit(qWidgetFactory):
    widget = qWidgetFactory(FloatEdit)
    yield widget


@pytest.fixture
def floatEditHolder(qWidgetFactory, floatEdit):
    widget = qWidgetFactory(qt.QWidget)
    layout = qt.QHBoxLayout(widget)
    layout.addStretch()
    layout.addWidget(floatEdit)
    yield widget


def test_show(floatEdit):
    pass


def test_value(floatEdit):
    floatEdit.setValue(1.5)
    assert floatEdit.value() == 1.5


def test_no_widgetresize(floatEditHolder, floatEdit):
    floatEditHolder.resize(50, 50)
    floatEdit.setValue(123)
    a = floatEdit.width()
    floatEdit.setValue(123456789123456789.123456789123456789)
    b = floatEdit.width()
    assert b == a


def test_widgetresize(qapp_utils, floatEditHolder, floatEdit):
    floatEditHolder.resize(50, 50)
    floatEdit.setWidgetResizable(True)
    # Initial
    floatEdit.setValue(123)
    qapp_utils.qWait()
    a = floatEdit.width()
    # Grow
    floatEdit.setValue(123456789123456789.123456789123456789)
    qapp_utils.qWait()
    b = floatEdit.width()
    # Shrink
    floatEdit.setValue(123)
    qapp_utils.qWait()
    c = floatEdit.width()
    assert b > a
    assert a <= c < b



# /*##########################################################################
#
# Copyright (c) 2016-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Module testing silx.app.view"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "07/06/2023"


import os
import sys
import shutil
import logging
import subprocess
import pytest

from .. import main
from silx import __main__ as silx_main

_logger = logging.getLogger(__name__)


def test_help(qapp):
    # option -h must cause a raise SystemExit or a return 0
    try:
        parser = main.createParser()
        parser.parse_args(["compare", "--help"])
        result = 0
    except SystemExit as e:
        result = e.args[0]
    assert result == 0


def test_wrong_option(qapp):
    try:
        parser = main.createParser()
        parser.parse_args(["compare", "--foo"])
        assert False
    except SystemExit as e:
        result = e.args[0]
    assert result != 0


def test_wrong_file(qapp):
    try:
        parser = main.createParser()
        result = parser.parse_args(["compare", "__file.not.found__"])
        result = 0
    except SystemExit as e:
        result = e.args[0]
    assert result == 0


def _create_test_env():
    """
    Returns an associated environment with a working project.
    """
    env = dict((str(k), str(v)) for k, v in os.environ.items())
    env["PYTHONPATH"] = os.pathsep.join(sys.path)
    return env


@pytest.fixture
def execute_as_script(tmp_path):
    """Execute a command line.

    Log output as debug in case of bad return code.
    """

    def execute_as_script(filename, *args):
        env = _create_test_env()

        # Copy file to temporary dir to avoid import from current dir.
        script = os.path.join(tmp_path, "launcher.py")
        shutil.copyfile(filename, script)
        command_line = [sys.executable, script] + list(args)

        _logger.info("Execute: %s", " ".join(command_line))
        p = subprocess.Popen(
            command_line, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env
        )
        out, err = p.communicate()
        _logger.info("Return code: %d", p.returncode)
        try:
            out = out.decode("utf-8")
        except UnicodeError:
            pass
        try:
            err = err.decode("utf-8")
        except UnicodeError:
            pass

        if p.returncode != 0:
            _logger.error("stdout:")
            _logger.error("%s", out)
            _logger.error("stderr:")
            _logger.error("%s", err)
        else:
            _logger.debug("stdout:")
            _logger.debug("%s", out)
            _logger.debug("stderr:")
            _logger.debug("%s", err)
        assert p.returncode == 0

    return execute_as_script


def test_execute_compare_help(qapp, execute_as_script):
    """Test if the main module is well connected.

    Uses subprocess to avoid to parasite the current environment.
    """
    execute_as_script(main.__file__, "--help")


def test_execute_silx_compare_help(qapp, execute_as_script):
    """Test if the main module is well connected.

    Uses subprocess to avoid to parasite the current environment.
    """
    execute_as_script(silx_main.__file__, "view", "--help")



# /*##########################################################################
#
# Copyright (c) 2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

from __future__ import annotations

from typing import NamedTuple, Any, ValuesView
from silx.gui import qt


class ProgressItem(NamedTuple):
    """Item storing the state of a stacked progress item"""

    value: int
    """Progression of the item"""

    visible: bool
    """Is the item displayed"""

    color: qt.QColor
    """Color of the progress"""

    striped: bool
    """If true, apply a stripe color to the gradiant"""

    animated: bool
    """If true, the stripe is animated"""

    toolTip: str
    """Tool tip of this item"""

    userData: Any
    """Any user data"""


class _UndefinedType:
    pass


_Undefined = _UndefinedType()


class StackedProgressBar(qt.QProgressBar):
    """
    Multiple stacked progress bar in single component
    """

    def __init__(self, parent: qt.Qwidget | None = None):
        super().__init__(parent=parent)
        self.__stacks: dict[str, ProgressItem] = {}
        self._animated: int = 0
        self._timer = qt.QTimer(self)
        self._timer.setInterval(80)
        self._timer.timeout.connect(self._tick)
        self._spacing: int = 0
        self._spacingCollapsible: bool = True

    def _tick(self):
        self._animated += 2
        self.update()

    def setSpacing(self, spacing: int):
        """Spacing between items, in pixels"""
        if self._spacing == spacing:
            return
        self._spacing = spacing
        self.update()

    def spacing(self) -> int:
        return self._spacing

    def setSpacingCollapsible(self, collapse: bool):
        """
        Set whether consecutive spacing should be collapsed.

        It can be useful to disable that to ensure pixel perfect
        rendering is some use cases.


        By default, this property is true.
        """
        if self._spacingCollapsible == collapse:
            return
        self._spacingCollapsible = collapse
        self.update()

    def spacingCollapsible(self) -> bool:
        return self._spacingCollapsible

    def clear(self):
        """Remove every stacked items from the widget"""
        if len(self.__stacks) == 0:
            return
        self.__stacks.clear()
        self.update()

    def setProgressItem(
        self,
        name: str,
        value: int | None | _UndefinedType = _Undefined,
        visible: bool | _UndefinedType = _Undefined,
        color: qt.QColor | None | _UndefinedType = _Undefined,
        striped: bool | _UndefinedType = _Undefined,
        animated: bool | _UndefinedType = _Undefined,
        toolTip: str | None | _UndefinedType = _Undefined,
        userData: Any = _Undefined,
    ):
        """Add or update a stacked items by its name"""

        previousItem = self.__stacks.get(name)

        if previousItem is not None:
            if value is _Undefined:
                value = previousItem.value
            if visible is _Undefined:
                visible = previousItem.visible
            if striped is _Undefined:
                striped = previousItem.striped
            if color is _Undefined:
                color = previousItem.color
            if toolTip is _Undefined:
                toolTip = previousItem.toolTip
            if animated is _Undefined:
                animated = previousItem.animated
            if userData is _Undefined:
                userData = previousItem.userData
        else:
            if value is _Undefined:
                value = 0
            if visible is _Undefined:
                visible = True
            if striped is _Undefined:
                striped = False
            if color is _Undefined:
                color = qt.QColor()
            if toolTip is _Undefined:
                toolTip = ""
            if animated is _Undefined:
                animated = False
            if userData is _Undefined:
                userData = None

        newItem = ProgressItem(
            value=value,
            visible=visible,
            color=color,
            striped=striped,
            animated=animated,
            toolTip=toolTip,
            userData=userData,
        )
        if previousItem == newItem:
            return
        self.__stacks[name] = newItem
        animated = any([s.animated for s in self.__stacks.values()])
        self._setAnimated(animated)
        self.update()

    def _setAnimated(self, animated: bool):
        if animated == self._timer.isActive():
            return
        if animated:
            self._timer.start()
        else:
            self._timer.stop()

    def removeProgressItem(self, name: str):
        """Remove a stacked item by its name"""
        s = self.__stacks.pop(name, None)
        if s is None:
            return
        self.update()

    def _brushFromProgressItem(self, item: ProgressItem) -> qt.QPalette | None:
        if item.color is None:
            return None

        palette = qt.QPalette()
        color = qt.QColor(item.color)

        if item.striped:
            if item.animated:
                delta = self._animated
            else:
                delta = 0
            color2 = color.lighter(120)
            shadowGradient = qt.QLinearGradient()
            shadowGradient.setSpread(qt.QGradient.RepeatSpread)
            shadowGradient.setStart(-delta, 0)
            shadowGradient.setFinalStop(8 - delta, -8)
            shadowGradient.setColorAt(0.0, color)
            shadowGradient.setColorAt(0.5, color)
            shadowGradient.setColorAt(0.50001, color2)
            shadowGradient.setColorAt(1.0, color2)
            brush = qt.QBrush(shadowGradient)
            palette.setBrush(qt.QPalette.Highlight, brush)
            palette.setBrush(qt.QPalette.Window, color2)
        else:
            palette.setColor(qt.QPalette.Highlight, color)

        return palette

    def paintEvent(self, event):
        painter = qt.QStylePainter(self)
        opt = qt.QStyleOptionProgressBar()
        self.initStyleOption(opt)
        painter.drawControl(qt.QStyle.CE_ProgressBarGroove, opt)
        self._drawProgressItems(painter, self.__stacks.values())

    def _drawProgressItems(self, painter: qt.QPainter, items: ValuesView[ProgressItem]):
        opt = qt.QStyleOptionProgressBar()
        self.initStyleOption(opt)

        visibleItems = [i for i in items if i.value and i.visible]
        xpos: int = 0
        w = opt.rect.width()
        if self._spacingCollapsible:
            cumspacing = max(0, len(visibleItems) - 1) * self._spacing
            w -= cumspacing
        vw = opt.maximum - opt.minimum
        opt.minimum = 0
        opt.maximum = w

        for item in visibleItems:
            xwidth = int(item.value * w / vw)
            opt.progress = xwidth * 2
            palette = self._brushFromProgressItem(item)
            if palette is not None:
                opt.palette = palette
            self._drawProgressItem(painter, opt, xpos, xwidth)
            xpos += xwidth + self._spacing

    def _drawProgressItem(
        self,
        painter: qt.QPainter,
        option: qt.QStyleOptionProgressBar,
        xpos: int,
        xwidth: int,
    ):
        if xwidth == 0:
            return
        rect: qt.QRect = option.rect
        style = self.style()

        if option.minimum == 0 and option.maximum == 0:
            return
        x0 = rect.x() + 3
        y0 = rect.y()

        h = rect.height()
        w = rect.width()
        xmaxwith = min(x0 + xpos + xwidth, w - 1) - x0 - xpos
        if xmaxwith < 0:
            return
        rect = qt.QRect(x0 + xpos, y0, xmaxwith, h)
        opt = qt.QStyleOptionProgressBar()
        opt.state = qt.QStyle.State_None
        margin = 1
        opt.rect = rect.marginsAdded(qt.QMargins(margin, margin, margin, margin))
        opt.palette = option.palette
        style.drawPrimitive(qt.QStyle.PE_IndicatorProgressChunk, opt, painter, self)

    def getProgressItemByPosition(self, pos: qt.QPoint) -> ProgressItem | None:
        """Returns the stacked item at a position of the component."""
        minimum = self.minimum()
        maximum = self.maximum()
        vRange = maximum - minimum
        w = self.width()
        v = pos.x() * vRange / w
        current = 0
        for item in self.__stacks.values():
            if not item.visible:
                continue
            current += item.value
            if v < current:
                return item
        return None

    def tooltipFromProgressItem(self, item: ProgressItem) -> str | None:
        """Returns the tooltip to display over an item.

        It is triggered when the tooltip have to be displayed.
        """
        return item.toolTip

    def event(self, event: qt.QEvent):
        if event.type() == qt.QEvent.ToolTip:
            item = self.getProgressItemByPosition(event.pos())
            if item is not None:
                toolTip = self.tooltipFromProgressItem(item)
                if toolTip:
                    qt.QToolTip.showText(event.globalPos(), toolTip, self)
            return True
        return super().event(event)



# /*##########################################################################
# Copyright (C) 2022-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Provides a wrapper to expose a dataset slice as a `commonh5.Dataset`."""

from __future__ import annotations

from typing import Tuple, Union

import h5py
import numpy

from . import commonh5
from . import utils


IndexType = Union[int, slice, type(Ellipsis)]
IndicesType = Union[IndexType, Tuple[IndexType, ...]]
NormalisedIndicesType = Tuple[Union[int, slice], ...]


def _expand_indices(
    ndim: int,
    indices: IndicesType,
) -> NormalisedIndicesType:
    """Replace Ellipsis and complete indices to match ndim"""
    if not isinstance(indices, tuple):
        indices = (indices,)

    nb_ellipsis = indices.count(Ellipsis)
    if nb_ellipsis > 1:
        raise IndexError("an index can only have a single ellipsis ('...')")
    if nb_ellipsis == 1:
        ellipsis_index = indices.index(Ellipsis)
        return (
            indices[:ellipsis_index]
            + (slice(None),) * max(0, (ndim - len(indices) + 1))
            + indices[ellipsis_index + 1 :]
        )

    if len(indices) > ndim:
        raise IndexError(
            f"too many indices ({len(indices)}) for the number of dimensions ({ndim})"
        )
    return indices + (slice(None),) * (ndim - len(indices))


def _get_selection_shape(
    shape: tuple[int, ...],
    indices: NormalisedIndicesType,
) -> tuple[int, ...]:
    """Returns the shape of the selection of indices in a dataset of the given shape"""
    assert len(shape) == len(indices)

    selected_indices = (
        index.indices(length)
        for length, index in zip(shape, indices)
        if isinstance(index, slice)
    )
    return tuple(
        int(max(0, numpy.ceil((stop - start) / stride)))
        for start, stop, stride in selected_indices
    )


def _combine_indices(
    outer_shape: tuple[int, ...],
    outer_indices: NormalisedIndicesType,
    indices: IndicesType,
) -> NormalisedIndicesType:
    """Returns the combination of outer_indices and indices"""
    inner_shape = _get_selection_shape(outer_shape, outer_indices)
    inner_indices = _expand_indices(len(inner_shape), indices)
    inner_iter = zip(range(len(inner_shape)), inner_shape, inner_indices)

    combined_indices = []
    for outer_length, outer_index in zip(outer_shape, outer_indices):
        if isinstance(outer_index, int):
            combined_indices.append(outer_index)
            continue

        outer_start, outer_stop, outer_stride = outer_index.indices(outer_length)
        inner_axis, inner_length, inner_index = next(inner_iter)

        if isinstance(inner_index, int):
            if inner_index < -inner_length or inner_index >= inner_length:
                raise IndexError(
                    f"index {inner_index} is out of bounds for axis {inner_axis} with size {inner_length}"
                )
            index = outer_start + outer_stride * inner_index
            if inner_index < 0:
                index += outer_stride * inner_length
            combined_indices.append(index)
            continue

        inner_start, inner_stop, inner_stride = inner_index.indices(inner_length)
        combined_indices.append(
            slice(
                outer_start + outer_stride * inner_start,
                outer_start + outer_stride * inner_stop,
                outer_stride * inner_stride,
            )
        )

    return tuple(combined_indices)


class DatasetSlice(commonh5.Dataset):
    """Wrapper a dataset indexed selection as a commonh5.Dataset.
    :param h5file: h5py-like file containing the dataset
    :param dataset: h5py-like dataset from which to access a slice
    :param indices: The indexing to select
    :param attrs: dataset attributes
    """

    def __init__(
        self,
        dataset: Union[h5py.Dataset, commonh5.Dataset],
        indices: IndicesType,
        attrs: dict,
    ):
        if not utils.is_dataset(dataset):
            raise ValueError(f"Unsupported dataset '{dataset}'")

        self.__dataset = dataset
        self.__file = dataset.file  # Keep a ref on file to fix issue recovering it
        self.__indices = indices
        self.__expanded_indices = _expand_indices(len(self.__dataset.shape), indices)
        self.__shape = _get_selection_shape(
            self.__dataset.shape, self.__expanded_indices
        )
        super().__init__(
            self.__dataset.name, data=None, parent=self.__file, attrs=attrs
        )

    def _get_data(self) -> Union[h5py.Dataset, commonh5.Dataset]:
        # Give access to the underlying (h5py) dataset, not the selected data
        # All commonh5.Dataset methods using _get_data must be overridden
        return self.__dataset

    @property
    def dtype(self) -> numpy.dtype:
        return self.__dataset.dtype

    @property
    def shape(self) -> tuple[int, ...]:
        return self.__shape

    @property
    def size(self) -> int:
        return numpy.prod(self.shape)

    def __len__(self) -> int:
        return self.shape[0]

    def __getitem__(self, item):
        if item is Ellipsis:
            return numpy.asarray(self.__dataset[self.__expanded_indices])
        if item == ():
            return self.__dataset[self.__expanded_indices]

        if not self.__shape:
            raise IndexError("invalid index to scalar variable.")

        return self.__dataset[
            _combine_indices(
                self.__dataset.shape,
                self.__expanded_indices,
                item,
            )
        ]

    @property
    def value(self):
        return self[()]

    def __iter__(self):
        return self[()].__iter__()

    @property
    def file(self) -> Union[h5py.File, commonh5.File]:
        if isinstance(self.__file, h5py.File) and not self.__file.id:
            return None
        return self.__file

    @property
    def name(self) -> str:
        return self.basename

    @property
    def indices(self) -> IndicesType:
        return self.__indices

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """Close the file"""
        self.__file.close()



# /*##########################################################################
# Copyright (C) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
Nominal tests of the HistogramndLut function.
"""

import unittest

import numpy as np

from silx.math import HistogramndLut


def _get_bin_edges(histo_range, n_bins, n_dims):
    edges = []
    for i_dim in range(n_dims):
        edges.append(
            histo_range[i_dim, 0]
            + np.arange(n_bins[i_dim] + 1)
            * (histo_range[i_dim, 1] - histo_range[i_dim, 0])
            / n_bins[i_dim]
        )
    return tuple(edges)


# ==============================================================
# ==============================================================
# ==============================================================


class _TestHistogramndLut_nominal(unittest.TestCase):
    """
    Unit tests of the HistogramndLut class.
    """

    __test__ = False  # ignore abstract class

    ndims = None

    def setUp(self):
        ndims = self.ndims
        if ndims is None:
            self.skipTest("Abstract class")
        self.tested_dim = ndims - 1

        if ndims is None:
            raise ValueError("ndims class member not set.")

        sample = np.array([5.5, -3.3, 0.0, -0.5, 3.3, 8.8, -7.7, 6.0, -4.0])

        weights = np.array(
            [500.5, -300.3, 0.01, -0.5, 300.3, 800.8, -700.7, 600.6, -400.4]
        )

        n_elems = len(sample)

        if ndims == 1:
            shape = (n_elems,)
        else:
            shape = (n_elems, ndims)

        self.sample = np.zeros(shape=shape, dtype=sample.dtype)
        if ndims == 1:
            self.sample = sample
        else:
            self.sample[..., ndims - 1] = sample

        self.weights = weights

        # the tests are performed along one dimension,
        #   all the other bins indices along the other dimensions
        #   are expected to be 2
        # (e.g : when testing a 2D sample : [0, x] will go into
        # bin [2, y] because of the bin ranges [-2, 2] and n_bins = 4
        # for the first dimension)
        self.other_axes_index = 2
        self.histo_range = np.repeat([[-2.0, 2.0]], ndims, axis=0)
        self.histo_range[ndims - 1] = [-4.0, 6.0]

        self.n_bins = np.array([4] * ndims)
        self.n_bins[ndims - 1] = 5

        if ndims == 1:

            def fill_histo(h, v, dim, op=None):
                if op:
                    h[:] = op(h[:], v)
                else:
                    h[:] = v

            self.fill_histo = fill_histo
        else:

            def fill_histo(h, v, dim, op=None):
                idx = [self.other_axes_index] * len(h.shape)
                idx[dim] = slice(0, None)
                idx = tuple(idx)
                if op:
                    h[idx] = op(h[idx], v)
                else:
                    h[idx] = v

            self.fill_histo = fill_histo

    def test_nominal_bin_edges(self):
        instance = HistogramndLut(self.sample, self.histo_range, self.n_bins)

        bin_edges = instance.bins_edges

        expected_edges = _get_bin_edges(self.histo_range, self.n_bins, self.ndims)

        for i_edges, edges in enumerate(expected_edges):
            self.assertTrue(
                np.array_equal(bin_edges[i_edges], expected_edges[i_edges]),
                msg="Testing bin_edges for dim {0}" "".format(i_edges + 1),
            )

    def test_nominal_histo_range(self):
        instance = HistogramndLut(self.sample, self.histo_range, self.n_bins)

        histo_range = instance.histo_range

        self.assertTrue(np.array_equal(histo_range, self.histo_range))

    def test_nominal_last_bin_closed(self):
        instance = HistogramndLut(self.sample, self.histo_range, self.n_bins)

        last_bin_closed = instance.last_bin_closed

        self.assertEqual(last_bin_closed, False)

        instance = HistogramndLut(
            self.sample, self.histo_range, self.n_bins, last_bin_closed=True
        )

        last_bin_closed = instance.last_bin_closed

        self.assertEqual(last_bin_closed, True)

        instance = HistogramndLut(
            self.sample, self.histo_range, self.n_bins, last_bin_closed=False
        )

        last_bin_closed = instance.last_bin_closed

        self.assertEqual(last_bin_closed, False)

    def test_nominal_n_bins_array(self):
        test_n_bins = np.arange(self.ndims) + 10
        instance = HistogramndLut(self.sample, self.histo_range, test_n_bins)

        n_bins = instance.n_bins

        self.assertTrue(np.array_equal(test_n_bins, n_bins))

    def test_nominal_n_bins_scalar(self):
        test_n_bins = 10
        expected_n_bins = np.array([test_n_bins] * self.ndims)
        instance = HistogramndLut(self.sample, self.histo_range, test_n_bins)

        n_bins = instance.n_bins

        self.assertTrue(np.array_equal(expected_n_bins, n_bins))

    def test_nominal_histo_ref(self):
        """ """
        expected_h_tpl = np.array([2, 1, 1, 1, 1])
        expected_c_tpl = np.array([-700.7, -0.5, 0.01, 300.3, 500.5])

        expected_h = np.zeros(shape=self.n_bins, dtype=np.double)
        expected_c = np.zeros(shape=self.n_bins, dtype=np.double)

        self.fill_histo(expected_h, expected_h_tpl, self.ndims - 1)
        self.fill_histo(expected_c, expected_c_tpl, self.ndims - 1)

        instance = HistogramndLut(self.sample, self.histo_range, self.n_bins)

        instance.accumulate(self.weights)

        histo = instance.histo()
        w_histo = instance.weighted_histo()
        histo_ref = instance.histo(copy=False)
        w_histo_ref = instance.weighted_histo(copy=False)

        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))
        self.assertTrue(np.array_equal(histo_ref, expected_h))
        self.assertTrue(np.array_equal(w_histo_ref, expected_c))

        histo_ref[0, ...] = histo_ref[0, ...] + 10
        w_histo_ref[0, ...] = w_histo_ref[0, ...] + 20

        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))
        self.assertFalse(np.array_equal(histo_ref, expected_h))
        self.assertFalse(np.array_equal(w_histo_ref, expected_c))

        histo_2 = instance.histo()
        w_histo_2 = instance.weighted_histo()

        self.assertFalse(np.array_equal(histo_2, expected_h))
        self.assertFalse(np.array_equal(w_histo_2, expected_c))
        self.assertTrue(np.array_equal(histo_2, histo_ref))
        self.assertTrue(np.array_equal(w_histo_2, w_histo_ref))

    def test_nominal_accumulate_once(self):
        """ """
        expected_h_tpl = np.array([2, 1, 1, 1, 1])
        expected_c_tpl = np.array([-700.7, -0.5, 0.01, 300.3, 500.5])

        expected_h = np.zeros(shape=self.n_bins, dtype=np.double)
        expected_c = np.zeros(shape=self.n_bins, dtype=np.double)

        self.fill_histo(expected_h, expected_h_tpl, self.ndims - 1)
        self.fill_histo(expected_c, expected_c_tpl, self.ndims - 1)

        instance = HistogramndLut(self.sample, self.histo_range, self.n_bins)

        instance.accumulate(self.weights)

        histo = instance.histo()
        w_histo = instance.weighted_histo()

        self.assertEqual(w_histo.dtype, np.float64)
        self.assertEqual(histo.dtype, np.uint32)
        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))
        self.assertTrue(np.array_equal(instance.histo(), expected_h))
        self.assertTrue(np.array_equal(instance.weighted_histo(), expected_c))

    def test_nominal_accumulate_twice(self):
        """ """
        expected_h_tpl = np.array([2, 1, 1, 1, 1])
        expected_c_tpl = np.array([-700.7, -0.5, 0.01, 300.3, 500.5])

        expected_h = np.zeros(shape=self.n_bins, dtype=np.double)
        expected_c = np.zeros(shape=self.n_bins, dtype=np.double)

        self.fill_histo(expected_h, expected_h_tpl, self.ndims - 1)
        self.fill_histo(expected_c, expected_c_tpl, self.ndims - 1)

        # calling accumulate twice
        expected_h *= 2
        expected_c *= 2

        instance = HistogramndLut(self.sample, self.histo_range, self.n_bins)

        instance.accumulate(self.weights)

        instance.accumulate(self.weights)

        histo = instance.histo()
        w_histo = instance.weighted_histo()

        self.assertEqual(w_histo.dtype, np.float64)
        self.assertEqual(histo.dtype, np.uint32)
        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))
        self.assertTrue(np.array_equal(instance.histo(), expected_h))
        self.assertTrue(np.array_equal(instance.weighted_histo(), expected_c))

    def test_nominal_apply_lut_once(self):
        """ """
        expected_h_tpl = np.array([2, 1, 1, 1, 1])
        expected_c_tpl = np.array([-700.7, -0.5, 0.01, 300.3, 500.5])

        expected_h = np.zeros(shape=self.n_bins, dtype=np.double)
        expected_c = np.zeros(shape=self.n_bins, dtype=np.double)

        self.fill_histo(expected_h, expected_h_tpl, self.ndims - 1)
        self.fill_histo(expected_c, expected_c_tpl, self.ndims - 1)

        instance = HistogramndLut(self.sample, self.histo_range, self.n_bins)

        histo, w_histo = instance.apply_lut(self.weights)

        self.assertEqual(w_histo.dtype, np.float64)
        self.assertEqual(histo.dtype, np.uint32)
        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))
        self.assertEqual(instance.histo(), None)
        self.assertEqual(instance.weighted_histo(), None)

    def test_nominal_apply_lut_twice(self):
        """ """
        expected_h_tpl = np.array([2, 1, 1, 1, 1])
        expected_c_tpl = np.array([-700.7, -0.5, 0.01, 300.3, 500.5])

        expected_h = np.zeros(shape=self.n_bins, dtype=np.double)
        expected_c = np.zeros(shape=self.n_bins, dtype=np.double)

        self.fill_histo(expected_h, expected_h_tpl, self.ndims - 1)
        self.fill_histo(expected_c, expected_c_tpl, self.ndims - 1)

        # calling apply_lut twice
        expected_h *= 2
        expected_c *= 2

        instance = HistogramndLut(self.sample, self.histo_range, self.n_bins)

        histo, w_histo = instance.apply_lut(self.weights)
        histo_2, w_histo_2 = instance.apply_lut(
            self.weights, histo=histo, weighted_histo=w_histo
        )

        self.assertEqual(id(histo), id(histo_2))
        self.assertEqual(id(w_histo), id(w_histo_2))
        self.assertEqual(w_histo.dtype, np.float64)
        self.assertEqual(histo.dtype, np.uint32)
        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))
        self.assertEqual(instance.histo(), None)
        self.assertEqual(instance.weighted_histo(), None)

    def test_nominal_accumulate_last_bin_closed(self):
        """ """
        expected_h_tpl = np.array([2, 1, 1, 1, 2])
        expected_c_tpl = np.array([-700.7, -0.5, 0.01, 300.3, 1101.1])

        expected_h = np.zeros(shape=self.n_bins, dtype=np.double)
        expected_c = np.zeros(shape=self.n_bins, dtype=np.double)

        self.fill_histo(expected_h, expected_h_tpl, self.ndims - 1)
        self.fill_histo(expected_c, expected_c_tpl, self.ndims - 1)

        instance = HistogramndLut(
            self.sample, self.histo_range, self.n_bins, last_bin_closed=True
        )

        instance.accumulate(self.weights)

        histo = instance.histo()
        w_histo = instance.weighted_histo()

        self.assertEqual(w_histo.dtype, np.float64)
        self.assertEqual(histo.dtype, np.uint32)
        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))

    def test_nominal_accumulate_weight_min_max(self):
        """ """
        weight_min = -299.9
        weight_max = 499.9

        expected_h_tpl = np.array([0, 1, 1, 1, 0])
        expected_c_tpl = np.array([0.0, -0.5, 0.01, 300.3, 0.0])

        expected_h = np.zeros(shape=self.n_bins, dtype=np.double)
        expected_c = np.zeros(shape=self.n_bins, dtype=np.double)

        self.fill_histo(expected_h, expected_h_tpl, self.ndims - 1)
        self.fill_histo(expected_c, expected_c_tpl, self.ndims - 1)

        instance = HistogramndLut(self.sample, self.histo_range, self.n_bins)

        instance.accumulate(self.weights, weight_min=weight_min, weight_max=weight_max)

        histo = instance.histo()
        w_histo = instance.weighted_histo()

        self.assertEqual(w_histo.dtype, np.float64)
        self.assertEqual(histo.dtype, np.uint32)
        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))

    def test_nominal_accumulate_forced_int32(self):
        """
        double weights, int32 weighted_histogram
        """
        expected_h_tpl = np.array([2, 1, 1, 1, 1])
        expected_c_tpl = np.array([-700, 0, 0, 300, 500])

        expected_h = np.zeros(shape=self.n_bins, dtype=np.double)
        expected_c = np.zeros(shape=self.n_bins, dtype=np.double)

        self.fill_histo(expected_h, expected_h_tpl, self.ndims - 1)
        self.fill_histo(expected_c, expected_c_tpl, self.ndims - 1)

        instance = HistogramndLut(
            self.sample, self.histo_range, self.n_bins, dtype=np.int32
        )

        instance.accumulate(self.weights)

        histo = instance.histo()
        w_histo = instance.weighted_histo()

        self.assertEqual(w_histo.dtype, np.int32)
        self.assertEqual(histo.dtype, np.uint32)
        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))

    def test_nominal_accumulate_forced_float32(self):
        """
        int32 weights, float32 weighted_histogram
        """
        expected_h_tpl = np.array([2, 1, 1, 1, 1])
        expected_c_tpl = np.array([-700.0, 0.0, 0.0, 300.0, 500.0])

        expected_h = np.zeros(shape=self.n_bins, dtype=np.double)
        expected_c = np.zeros(shape=self.n_bins, dtype=np.float32)

        self.fill_histo(expected_h, expected_h_tpl, self.ndims - 1)
        self.fill_histo(expected_c, expected_c_tpl, self.ndims - 1)

        instance = HistogramndLut(
            self.sample, self.histo_range, self.n_bins, dtype=np.float32
        )

        instance.accumulate(self.weights.astype(np.int32))

        histo = instance.histo()
        w_histo = instance.weighted_histo()

        self.assertEqual(w_histo.dtype, np.float32)
        self.assertEqual(histo.dtype, np.uint32)
        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))

    def test_nominal_accumulate_int32(self):
        """
        int32 weights
        """
        expected_h_tpl = np.array([2, 1, 1, 1, 1])
        expected_c_tpl = np.array([-700, 0, 0, 300, 500])

        expected_h = np.zeros(shape=self.n_bins, dtype=np.double)
        expected_c = np.zeros(shape=self.n_bins, dtype=np.int32)

        self.fill_histo(expected_h, expected_h_tpl, self.ndims - 1)
        self.fill_histo(expected_c, expected_c_tpl, self.ndims - 1)

        instance = HistogramndLut(self.sample, self.histo_range, self.n_bins)

        instance.accumulate(self.weights.astype(np.int32))

        histo = instance.histo()
        w_histo = instance.weighted_histo()

        self.assertEqual(w_histo.dtype, np.int32)
        self.assertEqual(histo.dtype, np.uint32)
        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))

    def test_nominal_accumulate_int32_double(self):
        """
        int32 weights
        """
        expected_h_tpl = np.array([2, 1, 1, 1, 1])
        expected_c_tpl = np.array([-700, 0, 0, 300, 500])

        expected_h = np.zeros(shape=self.n_bins, dtype=np.double)
        expected_c = np.zeros(shape=self.n_bins, dtype=np.int32)

        self.fill_histo(expected_h, expected_h_tpl, self.ndims - 1)
        self.fill_histo(expected_c, expected_c_tpl, self.ndims - 1)

        instance = HistogramndLut(self.sample, self.histo_range, self.n_bins)

        instance.accumulate(self.weights.astype(np.int32))
        instance.accumulate(self.weights)

        histo = instance.histo()
        w_histo = instance.weighted_histo()

        expected_h *= 2
        expected_c *= 2

        self.assertEqual(w_histo.dtype, np.int32)
        self.assertEqual(histo.dtype, np.uint32)
        self.assertTrue(np.array_equal(histo, expected_h))
        self.assertTrue(np.array_equal(w_histo, expected_c))

    def testNoneNativeTypes(self):
        type = self.sample.dtype.newbyteorder("B")
        sampleB = self.sample.astype(type)

        type = self.sample.dtype.newbyteorder("L")
        sampleL = self.sample.astype(type)

        histo_inst = HistogramndLut(sampleB, self.histo_range, self.n_bins)

        histo_inst = HistogramndLut(sampleL, self.histo_range, self.n_bins)


class TestHistogramndLut_nominal_1d(_TestHistogramndLut_nominal):
    __test__ = True  # because _TestHistogramndLut_nominal is ignored
    ndims = 1


class TestHistogramndLut_nominal_2d(_TestHistogramndLut_nominal):
    __test__ = True  # because _TestHistogramndLut_nominal is ignored
    ndims = 2


class TestHistogramndLut_nominal_3d(_TestHistogramndLut_nominal):
    __test__ = True  # because _TestHistogramndLut_nominal is ignored
    ndims = 3



# /*##########################################################################
#
# Copyright (c) 2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Test PlotWidget active item"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "11/12/2023"


import numpy
import pytest

from silx.gui.utils.testutils import SignalListener
from silx.gui.plot.items.curve import CurveStyle


@pytest.mark.parametrize("plotWidget", ("mpl", "gl"), indirect=True)
def testActiveCurveAndLabels(plotWidget):
    # Active curve handling off, no label change
    plotWidget.setActiveCurveHandling(False)
    plotWidget.getXAxis().setLabel("XLabel")
    plotWidget.getYAxis().setLabel("YLabel")
    plotWidget.addCurve((1, 2), (1, 2))
    assert plotWidget.getXAxis().getLabel() == "XLabel"
    assert plotWidget.getYAxis().getLabel() == "YLabel"

    plotWidget.addCurve((1, 2), (2, 3), xlabel="x1", ylabel="y1")
    assert plotWidget.getXAxis().getLabel() == "XLabel"
    assert plotWidget.getYAxis().getLabel() == "YLabel"

    plotWidget.clear()
    assert plotWidget.getXAxis().getLabel() == "XLabel"
    assert plotWidget.getYAxis().getLabel() == "YLabel"

    # Active curve handling on, label changes
    plotWidget.setActiveCurveHandling(True)
    plotWidget.getXAxis().setLabel("XLabel")
    plotWidget.getYAxis().setLabel("YLabel")

    # labels changed as active curve
    plotWidget.addCurve((1, 2), (1, 2), legend="1", xlabel="x1", ylabel="y1")
    plotWidget.setActiveCurve("1")
    assert plotWidget.getXAxis().getLabel() == "x1"
    assert plotWidget.getYAxis().getLabel() == "y1"

    # labels not changed as not active curve
    plotWidget.addCurve((1, 2), (2, 3), legend="2")
    assert plotWidget.getXAxis().getLabel() == "x1"
    assert plotWidget.getYAxis().getLabel() == "y1"

    # labels changed
    plotWidget.setActiveCurve("2")
    assert plotWidget.getXAxis().getLabel() == "XLabel"
    assert plotWidget.getYAxis().getLabel() == "YLabel"

    plotWidget.setActiveCurve("1")
    assert plotWidget.getXAxis().getLabel() == "x1"
    assert plotWidget.getYAxis().getLabel() == "y1"

    plotWidget.clear()
    assert plotWidget.getXAxis().getLabel() == "XLabel"
    assert plotWidget.getYAxis().getLabel() == "YLabel"

    plotWidget.setActiveCurveHandling(False)


@pytest.mark.parametrize("plotWidget", ("mpl", "gl"), indirect=True)
def testPlotActiveCurveSelectionMode(plotWidget):
    xData = numpy.arange(1000)
    yData = -500 + 100 * numpy.sin(xData)
    xData2 = xData + 1000
    yData2 = xData - 1000 + 200 * numpy.random.random(1000)

    plotWidget.clear()
    plotWidget.setActiveCurveHandling(True)
    legend = "curve 1"
    plotWidget.addCurve(xData, yData, legend=legend, color="green")

    # active curve should be None
    assert plotWidget.getActiveCurve(just_legend=True) is None

    # active curve should be None when None is set as active curve
    plotWidget.setActiveCurve(legend)
    current = plotWidget.getActiveCurve(just_legend=True)
    assert current == legend
    plotWidget.setActiveCurve(None)
    current = plotWidget.getActiveCurve(just_legend=True)
    assert current is None

    # testing it automatically toggles if there is only one
    plotWidget.setActiveCurveSelectionMode("legacy")
    current = plotWidget.getActiveCurve(just_legend=True)
    assert current == legend

    # active curve should not change when None set as active curve
    assert plotWidget.getActiveCurveSelectionMode() == "legacy"
    plotWidget.setActiveCurve(None)
    current = plotWidget.getActiveCurve(just_legend=True)
    assert current == legend

    # situation where no curve is active
    plotWidget.clear()
    plotWidget.setActiveCurveHandling(True)
    assert plotWidget.getActiveCurveSelectionMode() == "atmostone"
    plotWidget.addCurve(xData, yData, legend=legend, color="green")
    assert plotWidget.getActiveCurve(just_legend=True) is None
    plotWidget.addCurve(xData2, yData2, legend="curve 2", color="red")
    assert plotWidget.getActiveCurve(just_legend=True) is None
    plotWidget.setActiveCurveSelectionMode("legacy")
    assert plotWidget.getActiveCurve(just_legend=True) is None

    # the first curve added should be active
    plotWidget.clear()
    plotWidget.addCurve(xData, yData, legend=legend, color="green")
    assert plotWidget.getActiveCurve(just_legend=True) == legend
    plotWidget.addCurve(xData2, yData2, legend="curve 2", color="red")
    assert plotWidget.getActiveCurve(just_legend=True) == legend

    plotWidget.setActiveCurveHandling(False)


@pytest.mark.parametrize("plotWidget", ("mpl", "gl"), indirect=True)
def testActiveCurveStyle(plotWidget):
    """Test change of active curve style"""
    plotWidget.setActiveCurveHandling(True)
    plotWidget.setActiveCurveStyle(color="black")
    style = plotWidget.getActiveCurveStyle()
    assert style.getColor() == (0.0, 0.0, 0.0, 1.0)
    assert style.getLineStyle() is None
    assert style.getLineWidth() is None
    assert style.getSymbol() is None
    assert style.getSymbolSize() is None

    xData = numpy.arange(1000)
    yData = -500 + 100 * numpy.sin(xData)
    plotWidget.addCurve(x=xData, y=yData, legend="curve1")
    curve = plotWidget.getCurve("curve1")
    curve.setColor("blue")
    curve.setLineStyle("-")
    curve.setLineWidth(1)
    curve.setSymbol("o")
    curve.setSymbolSize(5)

    # Check default current style
    defaultStyle = curve.getCurrentStyle()
    assert defaultStyle == CurveStyle(
        color="blue", linestyle="-", linewidth=1, symbol="o", symbolsize=5
    )

    # Activate curve with highlight color=black
    plotWidget.setActiveCurve("curve1")
    style = curve.getCurrentStyle()
    assert style.getColor() == (0.0, 0.0, 0.0, 1.0)
    assert style.getLineStyle() == "-"
    assert style.getLineWidth() == 1
    assert style.getSymbol() == "o"
    assert style.getSymbolSize() == 5

    # Change highlight to linewidth=2
    plotWidget.setActiveCurveStyle(linewidth=2)
    style = curve.getCurrentStyle()
    assert style.getColor() == (0.0, 0.0, 1.0, 1.0)
    assert style.getLineStyle() == "-"
    assert style.getLineWidth() == 2
    assert style.getSymbol() == "o"
    assert style.getSymbolSize() == 5

    plotWidget.setActiveCurve(None)
    assert curve.getCurrentStyle() == defaultStyle

    plotWidget.setActiveCurveHandling(False)


@pytest.mark.parametrize("plotWidget", ("mpl", "gl"), indirect=True)
def testActiveImageAndLabels(plotWidget):
    # Active image handling always on, no API for toggling it
    plotWidget.getXAxis().setLabel("XLabel")
    plotWidget.getYAxis().setLabel("YLabel")

    # labels changed as active curve
    plotWidget.addImage(
        numpy.arange(100).reshape(10, 10), legend="1", xlabel="x1", ylabel="y1"
    )
    assert plotWidget.getXAxis().getLabel() == "x1"
    assert plotWidget.getYAxis().getLabel() == "y1"

    # labels not changed as not active curve
    plotWidget.addImage(numpy.arange(100).reshape(10, 10), legend="2")
    assert plotWidget.getXAxis().getLabel() == "x1"
    assert plotWidget.getYAxis().getLabel() == "y1"

    # labels changed
    plotWidget.setActiveImage("2")
    assert plotWidget.getXAxis().getLabel() == "XLabel"
    assert plotWidget.getYAxis().getLabel() == "YLabel"

    plotWidget.setActiveImage("1")
    assert plotWidget.getXAxis().getLabel() == "x1"
    assert plotWidget.getYAxis().getLabel() == "y1"

    plotWidget.clear()
    assert plotWidget.getXAxis().getLabel() == "XLabel"
    assert plotWidget.getYAxis().getLabel() == "YLabel"

    plotWidget.setActiveCurveHandling(False)


def _checkSelection(selection, current=None, selected=()):
    """Check current item and selected items."""
    assert selection.getCurrentItem() is current
    assert selection.getSelectedItems() == selected


@pytest.mark.parametrize("plotWidget", ("mpl", "gl"), indirect=True)
def testSelectionSyncWithActiveItems(plotWidget):
    """Test update of PlotWidgetSelection according to active items"""
    listener = SignalListener()

    selection = plotWidget.selection()
    selection.sigCurrentItemChanged.connect(listener)
    _checkSelection(selection)

    # Active item is current
    plotWidget.addImage(((0, 1), (2, 3)), legend="image")
    image = plotWidget.getActiveImage()
    assert listener.callCount() == 1
    _checkSelection(selection, image, (image,))

    # No active = no current
    plotWidget.setActiveImage(None)
    assert listener.callCount() == 2
    _checkSelection(selection)

    # Active item is current
    plotWidget.setActiveImage("image")
    assert listener.callCount() == 3
    _checkSelection(selection, image, (image,))

    # Mosted recently "actived" item is current
    plotWidget.addScatter((3, 2, 1), (0, 1, 2), (0, 1, 2), legend="scatter")
    scatter = plotWidget.getActiveScatter()
    assert listener.callCount() == 4
    _checkSelection(selection, scatter, (scatter, image))

    # Previously mosted recently "actived" item is current
    plotWidget.setActiveScatter(None)
    assert listener.callCount() == 5
    _checkSelection(selection, image, (image,))

    # Mosted recently "actived" item is current
    plotWidget.setActiveScatter("scatter")
    assert listener.callCount() == 6
    _checkSelection(selection, scatter, (scatter, image))

    # No active = no current
    plotWidget.setActiveImage(None)
    plotWidget.setActiveScatter(None)
    assert listener.callCount() == 7
    _checkSelection(selection)

    # Mosted recently "actived" item is current
    plotWidget.setActiveScatter("scatter")
    assert listener.callCount() == 8
    plotWidget.setActiveImage("image")
    assert listener.callCount() == 9
    _checkSelection(selection, image, (image, scatter))

    # Add a curve which is not active by default
    plotWidget.addCurve((0, 1, 2), (0, 1, 2), legend="curve")
    curve = plotWidget.getCurve("curve")
    assert listener.callCount() == 9
    _checkSelection(selection, image, (image, scatter))

    # Mosted recently "actived" item is current
    plotWidget.setActiveCurve("curve")
    assert listener.callCount() == 10
    _checkSelection(selection, curve, (curve, image, scatter))

    # Add a curve which is not active by default
    plotWidget.addCurve((0, 1, 2), (0, 1, 2), legend="curve2")
    curve2 = plotWidget.getCurve("curve2")
    assert listener.callCount() == 10
    _checkSelection(selection, curve, (curve, image, scatter))

    # Mosted recently "actived" item is current, previous curve is removed
    plotWidget.setActiveCurve("curve2")
    assert listener.callCount() == 11
    _checkSelection(selection, curve2, (curve2, image, scatter))

    # No items = no current
    plotWidget.clear()
    assert listener.callCount() == 12
    _checkSelection(selection)


@pytest.mark.parametrize("plotWidget", ("mpl", "gl"), indirect=True)
def testSelectionWithItems(plotWidget):
    """Test init of selection on a plot with items"""
    plotWidget.addImage(((0, 1), (2, 3)), legend="image")
    plotWidget.addScatter((3, 2, 1), (0, 1, 2), (0, 1, 2), legend="scatter")
    plotWidget.addCurve((0, 1, 2), (0, 1, 2), legend="curve")
    plotWidget.setActiveCurve("curve")

    selection = plotWidget.selection()
    assert selection.getCurrentItem() is not None
    selected = selection.getSelectedItems()
    assert len(selected) == 3
    assert plotWidget.getActiveCurve() in selected
    assert plotWidget.getActiveImage() in selected
    assert plotWidget.getActiveScatter() in selected


@pytest.mark.parametrize("plotWidget", ("mpl", "gl"), indirect=True)
def testSelectionSetCurrentItem(plotWidget):
    """Test setCurrentItem"""
    # Add items to the plot
    plotWidget.addImage(((0, 1), (2, 3)), legend="image")
    image = plotWidget.getActiveImage()
    plotWidget.addScatter((3, 2, 1), (0, 1, 2), (0, 1, 2), legend="scatter")
    scatter = plotWidget.getActiveScatter()
    plotWidget.addCurve((0, 1, 2), (0, 1, 2), legend="curve")
    plotWidget.setActiveCurve("curve")
    curve = plotWidget.getActiveCurve()

    selection = plotWidget.selection()
    assert selection.getCurrentItem() is not None
    assert len(selection.getSelectedItems()) == 3

    # Set current to None reset all active items
    selection.setCurrentItem(None)
    _checkSelection(selection)
    assert plotWidget.getActiveCurve() is None
    assert plotWidget.getActiveImage() is None
    assert plotWidget.getActiveScatter() is None

    # Set current to an item makes it active
    selection.setCurrentItem(image)
    _checkSelection(selection, image, (image,))
    assert plotWidget.getActiveCurve() is None
    assert plotWidget.getActiveImage() is image
    assert plotWidget.getActiveScatter() is None

    # Set current to an item makes it active and keeps other active
    selection.setCurrentItem(curve)
    _checkSelection(selection, curve, (curve, image))
    assert plotWidget.getActiveCurve() is curve
    assert plotWidget.getActiveImage() is image
    assert plotWidget.getActiveScatter() is None

    # Set current to an item makes it active and keeps other active
    selection.setCurrentItem(scatter)
    _checkSelection(selection, scatter, (scatter, curve, image))
    assert plotWidget.getActiveCurve() is curve
    assert plotWidget.getActiveImage() is image
    assert plotWidget.getActiveScatter() is scatter


def testSetActiveCurveWithInstance(plotWidget):
    """Test setting the active curve with a curve item instance"""
    plotWidget.addCurve((0, 1), (0, 1), legend="curve0")
    plotWidget.addCurve((0, 1), (1, 0), legend="curve1")
    curve0, curve1 = plotWidget.getItems()

    plotWidget.setActiveCurve(curve0)
    assert plotWidget.getActiveCurve() is curve0

    plotWidget.setActiveCurve(curve1)
    assert plotWidget.getActiveCurve() is curve1

    plotWidget.setActiveCurve(None)
    assert plotWidget.getActiveCurve() is None


def testSetActiveImageWithInstance(plotWidget):
    """Test setting the active image with an image item instance"""
    plotWidget.addImage(((0, 1), (2, 3)), legend="image")
    image = plotWidget.getItems()[0]

    plotWidget.setActiveImage(None)
    assert plotWidget.getActiveImage() is None

    plotWidget.setActiveImage(image)
    assert plotWidget.getActiveImage() is image


def testSetActiveScatterWithInstance(plotWidget):
    """Test setting the active scatter with a scatter item instance"""
    plotWidget.addScatter((0, 1), (0, 1), (0, 1), legend="scatter")
    scatter = plotWidget.getItems()[0]

    plotWidget.setActiveScatter(None)
    assert plotWidget.getActiveScatter() is None

    plotWidget.setActiveScatter(scatter)
    assert plotWidget.getActiveScatter() is scatter



# /*##########################################################################
#
# Copyright (c) 2017-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides markers item of the :class:`Plot`.
"""
from __future__ import annotations

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "06/03/2017"


import logging
import numpy

from ....utils.proxy import docstring
from .core import (
    Item,
    DraggableMixIn,
    ColorMixIn,
    LineMixIn,
    SymbolMixIn,
    ItemChangedType,
    YAxisMixIn,
)
from silx import config
from silx.gui import qt
from silx.gui import colors


_logger = logging.getLogger(__name__)


class MarkerBase(Item, DraggableMixIn, ColorMixIn, YAxisMixIn):
    """Base class for markers"""

    sigDragStarted = qt.Signal()
    """Signal emitted when the marker is pressed"""
    sigDragFinished = qt.Signal()
    """Signal emitted when the marker is released"""

    _DEFAULT_COLOR = (0.0, 0.0, 0.0, 1.0)
    """Default color of the markers"""

    def __init__(self):
        Item.__init__(self)
        DraggableMixIn.__init__(self)
        ColorMixIn.__init__(self)
        YAxisMixIn.__init__(self)

        self._text = ""
        self._font = None
        if config.DEFAULT_PLOT_MARKER_TEXT_FONT_SIZE is not None:
            self._font = qt.QFont(
                qt.QApplication.instance().font().family(),
                config.DEFAULT_PLOT_MARKER_TEXT_FONT_SIZE,
            )

        self._x = None
        self._y = None
        self._symbol_size = 10.0
        self._bgColor: colors.RGBAColorType | None = None
        self._constraint = self._defaultConstraint
        self.__isBeingDragged = False

    def _addRendererCall(
        self,
        backend,
        symbol=None,
        symbolsize=10.,
        linestyle="-",
        linewidth=1,
    ):
        """Perform the update of the backend renderer"""
        return backend.addMarker(
            x=self.getXPosition(),
            y=self.getYPosition(),
            text=self.getText(),
            color=self.getColor(),
            symbol=symbol,
            symbolsize=symbolsize,
            linestyle=linestyle,
            linewidth=linewidth,
            constraint=self.getConstraint(),
            yaxis=self.getYAxis(),
            font=self._font,  # Do not use getFont to spare creating a new QFont
            bgcolor=self.getBackgroundColor(),
        )

    def _addBackendRenderer(self, backend):
        """Update backend renderer"""
        raise NotImplementedError()

    @docstring(DraggableMixIn)
    def drag(self, from_, to):
        self.setPosition(to[0], to[1])

    def isOverlay(self):
        """Returns True: A marker is always rendered as an overlay.

        :rtype: bool
        """
        return True

    def getText(self):
        """Returns marker text.

        :rtype: str
        """
        return self._text

    def setText(self, text):
        """Set the text of the marker.

        :param str text: The text to use
        """
        text = str(text)
        if text != self._text:
            self._text = text
            self._updated(ItemChangedType.TEXT)

    def getFont(self) -> qt.QFont | None:
        """Returns a copy of the QFont used to render text.

        To modify the text font, use :meth:`setFont`.
        """
        return None if self._font is None else qt.QFont(self._font)

    def setFont(self, font: qt.QFont | None):
        """Set the QFont used to render text, use None for default.

        A copy is stored, so further modification of the provided font are not taken into account.
        """
        if font != self._font:
            self._font = None if font is None else qt.QFont(font)
            self._updated(ItemChangedType.FONT)

    def getBackgroundColor(self) -> colors.RGBAColorType | None:
        """Returns the RGBA background color of the item"""
        return self._bgColor

    def setBackgroundColor(self, color):
        """Set item text background color

        :param color: color(s) to be used as a str ("#RRGGBB") or (npoints, 4)
                      unsigned byte array or one of the predefined color names
                      defined in colors.py
        """
        if color is not None:
            color = colors.rgba(color)
        if self._bgColor != color:
            self._bgColor = color
            self._updated(ItemChangedType.BACKGROUND_COLOR)

    def getXPosition(self):
        """Returns the X position of the marker line in data coordinates

        :rtype: float or None
        """
        return self._x

    def getYPosition(self):
        """Returns the Y position of the marker line in data coordinates

        :rtype: float or None
        """
        return self._y

    def getPosition(self) -> tuple[float | None, float | None]:
        """Returns the (x, y) position of the marker in data coordinates

        :rtype: 2-tuple of float or None
        """
        return self._x, self._y

    def setPosition(self, x: float, y: float):
        """Set marker position in data coordinates

        Constraint are applied if any.

        :param float x: X coordinates in data frame
        :param float y: Y coordinates in data frame
        """
        x, y = self.getConstraint()(x, y)
        x, y = float(x), float(y)
        if x != self._x or y != self._y:
            self._x, self._y = x, y
            self._updated(ItemChangedType.POSITION)

    def getConstraint(self):
        """Returns the dragging constraint of this item"""
        return self._constraint

    def _setConstraint(self, constraint):  # TODO support update
        """Set the constraint.

        This is private for now as update is not handled.

        :param callable constraint:
        :param constraint: A function filtering item displacement by
                           dragging operations or None for no filter.
                           This function is called each time the item is
                           moved.
                           This is only used if isDraggable returns True.
        :type constraint: None or a callable that takes the coordinates of
                          the current cursor position in the plot as input
                          and that returns the filtered coordinates.
        """
        if constraint is None:
            constraint = self._defaultConstraint
        assert callable(constraint)
        self._constraint = constraint

    @staticmethod
    def _defaultConstraint(*args):
        """Default constraint not doing anything"""
        return args

    def _startDrag(self):
        self.__isBeingDragged = True
        self.sigDragStarted.emit()

    def _endDrag(self):
        self.__isBeingDragged = False
        self.sigDragFinished.emit()

    def isBeingDragged(self) -> bool:
        """Returns whether the marker is currently dragged by the user."""
        return self.__isBeingDragged


class Marker(MarkerBase, SymbolMixIn):
    """Description of a marker"""

    _DEFAULT_SYMBOL = "+"
    """Default symbol of the marker"""

    _DEFAULT_SYMBOL_SIZE = 10.0
    """Default size of marker's symbol"""

    def __init__(self):
        MarkerBase.__init__(self)
        SymbolMixIn.__init__(self)

        self._x = 0.0
        self._y = 0.0

    def _addBackendRenderer(self, backend):
        return self._addRendererCall(
            backend,
            symbol=self.getSymbol(),
            symbolsize=self.getSymbolSize(),
        )

    def _setConstraint(self, constraint):
        """Set the constraint function of the marker drag.

        It also supports 'horizontal' and 'vertical' str as constraint.

        :param constraint: The constraint of the dragging of this marker
        :type: constraint: callable or str
        """
        if constraint == "horizontal":
            constraint = self._horizontalConstraint
        elif constraint == "vertical":
            constraint = self._verticalConstraint

        super(Marker, self)._setConstraint(constraint)

    def _horizontalConstraint(self, _, y):
        return self.getXPosition(), y

    def _verticalConstraint(self, x, _):
        return x, self.getYPosition()


class _LineMarker(MarkerBase, LineMixIn):
    """Base class for line markers"""

    def __init__(self):
        MarkerBase.__init__(self)
        LineMixIn.__init__(self)

    def _addBackendRenderer(self, backend):
        return self._addRendererCall(
            backend, linestyle=self.getLineStyle(), linewidth=self.getLineWidth()
        )


class XMarker(_LineMarker):
    """Description of a marker"""

    def __init__(self):
        _LineMarker.__init__(self)
        self._x = 0.0

    def setPosition(self, x, y):
        """Set marker line position in data coordinates

        Constraint are applied if any.

        :param float x: X coordinates in data frame
        :param float y: Y coordinates in data frame
        """
        x, _ = self.getConstraint()(x, y)
        x = float(x)
        if x != self._x:
            self._x = x
            self._updated(ItemChangedType.POSITION)


class YMarker(_LineMarker):
    """Description of a marker"""

    def __init__(self):
        _LineMarker.__init__(self)
        self._y = 0.0

    def setPosition(self, x, y):
        """Set marker line position in data coordinates

        Constraint are applied if any.

        :param float x: X coordinates in data frame
        :param float y: Y coordinates in data frame
        """
        _, y = self.getConstraint()(x, y)
        y = float(y)
        if y != self._y:
            self._y = y
            self._updated(ItemChangedType.POSITION)



# /*##########################################################################
#
# Copyright (c) 2018-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "28/06/2018"


import pytest
import numpy.testing

import silx.gui.plot.items.roi as roi_items


def testLine_geometry(qapp):
    item = roi_items.LineROI()
    startPoint = numpy.array([1, 2])
    endPoint = numpy.array([3, 4])
    item.setEndPoints(startPoint, endPoint)
    numpy.testing.assert_allclose(item.getEndPoints()[0], startPoint)
    numpy.testing.assert_allclose(item.getEndPoints()[1], endPoint)


def testHLine_geometry(qapp):
    item = roi_items.HorizontalLineROI()
    item.setPosition(15)
    assert item.getPosition() == 15


def testVLine_geometry(qapp):
    item = roi_items.VerticalLineROI()
    item.setPosition(15)
    assert item.getPosition() == 15


def testPoint_geometry(qapp):
    point = numpy.array([1, 2])
    item = roi_items.PointROI()
    item.setPosition(point)
    numpy.testing.assert_allclose(item.getPosition(), point)


def testRectangle_originGeometry(qapp):
    origin = numpy.array([0, 0])
    size = numpy.array([10, 20])
    center = numpy.array([5, 10])
    item = roi_items.RectangleROI()
    item.setGeometry(origin=origin, size=size)
    numpy.testing.assert_allclose(item.getOrigin(), origin)
    numpy.testing.assert_allclose(item.getSize(), size)
    numpy.testing.assert_allclose(item.getCenter(), center)


def testRectangle_centerGeometry(qapp):
    origin = numpy.array([0, 0])
    size = numpy.array([10, 20])
    center = numpy.array([5, 10])
    item = roi_items.RectangleROI()
    item.setGeometry(center=center, size=size)
    numpy.testing.assert_allclose(item.getOrigin(), origin)
    numpy.testing.assert_allclose(item.getSize(), size)
    numpy.testing.assert_allclose(item.getCenter(), center)


def testRectangle_setCenterGeometry(qapp):
    origin = numpy.array([0, 0])
    size = numpy.array([10, 20])
    item = roi_items.RectangleROI()
    item.setGeometry(origin=origin, size=size)
    newCenter = numpy.array([0, 0])
    item.setCenter(newCenter)
    expectedOrigin = numpy.array([-5, -10])
    numpy.testing.assert_allclose(item.getOrigin(), expectedOrigin)
    numpy.testing.assert_allclose(item.getCenter(), newCenter)
    numpy.testing.assert_allclose(item.getSize(), size)


def testRectangle_setOriginGeometry(qapp):
    origin = numpy.array([0, 0])
    size = numpy.array([10, 20])
    item = roi_items.RectangleROI()
    item.setGeometry(origin=origin, size=size)
    newOrigin = numpy.array([10, 10])
    item.setOrigin(newOrigin)
    expectedCenter = numpy.array([15, 20])
    numpy.testing.assert_allclose(item.getOrigin(), newOrigin)
    numpy.testing.assert_allclose(item.getCenter(), expectedCenter)
    numpy.testing.assert_allclose(item.getSize(), size)


def testCircle_geometry(qapp):
    center = numpy.array([0, 0])
    radius = 10.0
    item = roi_items.CircleROI()
    item.setGeometry(center=center, radius=radius)
    numpy.testing.assert_allclose(item.getCenter(), center)
    numpy.testing.assert_allclose(item.getRadius(), radius)


def testCircle_setCenter(qapp):
    center = numpy.array([0, 0])
    radius = 10.0
    item = roi_items.CircleROI()
    item.setGeometry(center=center, radius=radius)
    newCenter = numpy.array([-10, 0])
    item.setCenter(newCenter)
    numpy.testing.assert_allclose(item.getCenter(), newCenter)
    numpy.testing.assert_allclose(item.getRadius(), radius)


def testCircle_setRadius(qapp):
    center = numpy.array([0, 0])
    radius = 10.0
    item = roi_items.CircleROI()
    item.setGeometry(center=center, radius=radius)
    newRadius = 5.1
    item.setRadius(newRadius)
    numpy.testing.assert_allclose(item.getCenter(), center)
    numpy.testing.assert_allclose(item.getRadius(), newRadius)


def testCircle_contains(qapp):
    center = numpy.array([2, -1])
    radius = 1.0
    item = roi_items.CircleROI()
    item.setGeometry(center=center, radius=radius)
    assert item.contains([1, -1])
    assert not item.contains([0, 0])
    assert item.contains([2, 0])
    assert not item.contains([3.01, -1])


def testEllipse_contains(qapp):
    center = numpy.array([-2, 0])
    item = roi_items.EllipseROI()
    item.setCenter(center)
    item.setOrientation(numpy.pi / 4.0)
    item.setMajorRadius(2)
    item.setMinorRadius(1)
    print(item.getMinorRadius(), item.getMajorRadius())
    assert not item.contains([0, 0])
    assert item.contains([-1, 1])
    assert item.contains([-3, 0])
    assert item.contains([-2, 0])
    assert item.contains([-2, 1])
    assert not item.contains([-4, 1])


def testRectangle_isIn(qapp):
    origin = numpy.array([0, 0])
    size = numpy.array([10, 20])
    item = roi_items.RectangleROI()
    item.setGeometry(origin=origin, size=size)
    assert item.contains(position=(0, 0))
    assert item.contains(position=(2, 14))
    assert not item.contains(position=(14, 12))


def testPolygon_emptyGeometry(qapp):
    points = numpy.empty((0, 2))
    item = roi_items.PolygonROI()
    item.setPoints(points)
    numpy.testing.assert_allclose(item.getPoints(), points)


def testPolygon_geometry(qapp):
    points = numpy.array([[10, 10], [12, 10], [50, 1]])
    item = roi_items.PolygonROI()
    item.setPoints(points)
    numpy.testing.assert_allclose(item.getPoints(), points)


def testPolygon_isIn(qapp):
    points = numpy.array([[0, 0], [0, 10], [5, 10]])
    item = roi_items.PolygonROI()
    item.setPoints(points)
    assert item.contains((0, 0))
    assert not item.contains((6, 2))
    assert not item.contains((-2, 5))
    assert not item.contains((2, -1))
    assert not item.contains((8, 1))
    assert item.contains((1, 8))


def testArc_getToSetGeometry(qapp):
    """Test that we can use getGeometry as input to setGeometry"""
    item = roi_items.ArcROI()
    item.setFirstShapePoints(numpy.array([[5, 10], [50, 100]]))
    item.setGeometry(*item.getGeometry())


def testArc_degenerated_point(qapp):
    item = roi_items.ArcROI()
    center = numpy.array([10, 20])
    innerRadius, outerRadius, startAngle, endAngle = 0, 0, 0, 0
    item.setGeometry(center, innerRadius, outerRadius, startAngle, endAngle)


def testArc_degenerated_line(qapp):
    item = roi_items.ArcROI()
    center = numpy.array([10, 20])
    innerRadius, outerRadius, startAngle, endAngle = 0, 100, numpy.pi, numpy.pi
    item.setGeometry(center, innerRadius, outerRadius, startAngle, endAngle)


def testArc_special_circle(qapp):
    item = roi_items.ArcROI()
    center = numpy.array([10, 20])
    innerRadius, outerRadius, startAngle, endAngle = 0, 100, numpy.pi, 3 * numpy.pi
    item.setGeometry(center, innerRadius, outerRadius, startAngle, endAngle)
    numpy.testing.assert_allclose(item.getCenter(), center)
    assert item.getInnerRadius() == pytest.approx(innerRadius)
    assert item.getOuterRadius() == pytest.approx(outerRadius)
    assert item.getStartAngle() == pytest.approx(item.getEndAngle() - numpy.pi * 2.0)
    assert item.isClosed()


def testArc_special_donut(qapp):
    item = roi_items.ArcROI()
    center = numpy.array([10, 20])
    innerRadius, outerRadius, startAngle, endAngle = 1, 100, numpy.pi, 3 * numpy.pi
    item.setGeometry(center, innerRadius, outerRadius, startAngle, endAngle)
    numpy.testing.assert_allclose(item.getCenter(), center)
    assert item.getInnerRadius() == pytest.approx(innerRadius)
    assert item.getOuterRadius() == pytest.approx(outerRadius)
    assert item.getStartAngle() == pytest.approx(item.getEndAngle() - numpy.pi * 2.0)
    assert item.isClosed()


def testArc_clockwiseGeometry(qapp):
    """Test that we can use getGeometry as input to setGeometry"""
    item = roi_items.ArcROI()
    center = numpy.array([10, 20])
    innerRadius, outerRadius, startAngle, endAngle = 1, 100, numpy.pi * 0.5, numpy.pi
    item.setGeometry(center, innerRadius, outerRadius, startAngle, endAngle)
    numpy.testing.assert_allclose(item.getCenter(), center)
    assert item.getInnerRadius() == pytest.approx(innerRadius)
    assert item.getOuterRadius() == pytest.approx(outerRadius)
    assert item.getStartAngle() == pytest.approx(startAngle)
    assert item.getEndAngle() == pytest.approx(endAngle)
    assert not item.isClosed()


def testArc_anticlockwiseGeometry(qapp):
    """Test that we can use getGeometry as input to setGeometry"""
    item = roi_items.ArcROI()
    center = numpy.array([10, 20])
    innerRadius, outerRadius, startAngle, endAngle = (
        1,
        100,
        numpy.pi * 0.5,
        -numpy.pi * 0.5,
    )
    item.setGeometry(center, innerRadius, outerRadius, startAngle, endAngle)
    numpy.testing.assert_allclose(item.getCenter(), center)
    assert item.getInnerRadius() == pytest.approx(innerRadius)
    assert item.getOuterRadius() == pytest.approx(outerRadius)
    assert item.getStartAngle() == pytest.approx(startAngle)
    assert item.getEndAngle() == pytest.approx(endAngle)
    assert not item.isClosed()


def testArc_position(qapp):
    """Test validity of getPosition"""
    item = roi_items.ArcROI()
    center = numpy.array([10, 20])
    innerRadius, outerRadius, startAngle, endAngle = 1, 100, numpy.pi * 0.5, numpy.pi
    item.setGeometry(center, innerRadius, outerRadius, startAngle, endAngle)
    assert item.getPosition(roi_items.ArcROI.Role.START) == pytest.approx((10.0, 70.5))
    assert item.getPosition(roi_items.ArcROI.Role.STOP) == pytest.approx((-40.5, 20.0))
    assert item.getPosition(roi_items.ArcROI.Role.MIDDLE) == pytest.approx(
        (-25.71, 55.71), abs=0.1
    )
    assert item.getPosition(roi_items.ArcROI.Role.CENTER) == pytest.approx(
        (10.0, 20), abs=0.1
    )


def testHRange_geometry(qapp):
    item = roi_items.HorizontalRangeROI()
    vmin = 1
    vmax = 3
    item.setRange(vmin, vmax)
    assert item.getMin() == pytest.approx(vmin)
    assert item.getMax() == pytest.approx(vmax)
    assert item.getCenter() == pytest.approx(2)


def testBand_getToSetGeometry(qapp):
    """Test that we can use getGeometry as input to setGeometry"""
    item = roi_items.BandROI()
    item.setFirstShapePoints(numpy.array([[5, 10], [50, 100]]))
    item.setGeometry(*item.getGeometry())



# /*##########################################################################
#
# Copyright (c) 2015-2024 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This package provides test of the root modules
"""

from __future__ import annotations

from collections.abc import Sequence
import importlib
import logging
import subprocess
import sys


try:
    import pytest
except ImportError:
    logging.getLogger(__name__).error(
        "pytest is required to run the tests, please install it."
    )
    raise


import silx


def run_tests(
    module: str | None = "silx",
    verbosity: int = 0,
    args: Sequence[str] = (),
):
    """Run tests in a subprocess

    :param module: Name of the silx module to test (default: 'silx')
    :param verbosity: Requested level of verbosity
    :param args: List of extra arguments to pass to `pytest`
    """
    cmd = [
        sys.executable,
        "-m",
        "pytest",
        f"--rootdir={silx.__path__[0]}",
        "--verbosity",
        str(verbosity),
        # Handle warning as errors unless explicitly skipped
        "-Werror",
        "-Wignore:tostring() is deprecated. Use tobytes() instead.:DeprecationWarning:OpenGL.GL.VERSION.GL_2_0",
        "-Wignore:Jupyter is migrating its paths to use standard platformdirs:DeprecationWarning",
        "-Wignore:Unable to import recommended hash 'siphash24.siphash13', falling back to 'hashlib.sha256'. Run 'python3 -m pip install siphash24' to install the recommended hash.:UserWarning:pytools.persistent_dict",
        "-Wignore:Non-empty compiler output encountered. Set the environment variable PYOPENCL_COMPILER_OUTPUT=1 to see more.:UserWarning",
        # Remove __array__ ignore once h5py v3.12 is released
        "-Wignore:__array__ implementation doesn't accept a copy keyword, so passing copy=False failed. __array__ must implement 'dtype' and 'copy' keyword arguments.:DeprecationWarning",
    ] + list(args)

    if module is not None:
        # Retrieve folder for packages and file for modules
        imported_module = importlib.import_module(module)
        cmd.append(
            imported_module.__path__[0]
            if hasattr(imported_module, "__path__")
            else imported_module.__file__
        )

    return subprocess.run(cmd, check=False).returncode
